#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib.curves import secp256k1 as ec
from btclib.der import _deserialize, _serialize


class TestDER(unittest.TestCase):
    def test_der(self):

        sighash_all = b"\x01"

        sig9 = 1, 1
        sig73 = ec.n - 1, ec.n - 1
        sig72 = 2 ** 255 - 1, ec.n - 1
        sig71 = 2 ** 255 - 1, 2 ** 255 - 1
        sig71b = 2 ** 255 - 1, 2 ** 248 - 1
        sig70 = 2 ** 255 - 1, 2 ** 247 - 1
        sig69 = 2 ** 247 - 1, 2 ** 247 - 1
        sigs = [sig9, sig73, sig72, sig71, sig71b, sig70, sig69]
        lenghts = [9, 73, 72, 71, 71, 70, 69]

        for lenght, sig in zip(lenghts, sigs):
            dersig = _serialize(*sig, sighash_all)
            r, s, sighash = _deserialize(dersig)
            self.assertEqual(sig, (r, s))
            self.assertEqual(sighash_all, sighash)
            self.assertEqual(len(dersig), lenght)
            # without sighash
            r, s, no_sighash = _deserialize(dersig[:-1])
            self.assertEqual(sig, (r, s))
            self.assertIsNone(no_sighash)

        # with the last one

        self.assertEqual((r, s, sighash), _deserialize((r, s, sighash)))

        # Invalid sighash (b'\x00')
        badsig = dersig[:-1] + b"\x00"
        self.assertRaises(ValueError, _deserialize, badsig)
        # _deserialize(badsig)

        # DER signature size should be in [8, 73]
        dersig2 = dersig + b"\x00" * 70
        self.assertRaises(ValueError, _deserialize, dersig2)

        # DER signature must be of type 0x30 (compound)
        dersig2 = b"\x00" + dersig[1:]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # Declared signature size does not match with size
        dersig2 = dersig[:1] + b"\x00" + dersig[2:]
        self.assertRaises(ValueError, _deserialize, dersig2)

        Rsize = dersig[3]
        # Zero-size integers are not allowed for r
        dersig2 = dersig[:3] + b"\x00" + dersig[4:]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # Length of the s scalar must be inside the signature
        dersig2 = dersig[:3] + b"\x80" + dersig[4:]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # Zero-size integers are not allowed for s
        dersig2 = dersig[: Rsize + 5] + b"\x00" + dersig[Rsize + 6 :]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # Signature size does not match with scalars
        dersig2 = dersig[: Rsize + 5] + b"\x4f" + dersig[Rsize + 6 :]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # r scalar must be an integer
        dersig2 = dersig[:2] + b"\x00" + dersig[3:]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # Negative numbers are not allowed for r
        dersig2 = dersig[:4] + b"\x80" + dersig[5:]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # Invalid null bytes at the start of r
        dersig2 = dersig[:4] + b"\x00\x00" + dersig[6:]
        self.assertRaises(ValueError, _deserialize, dersig2)
        # deserialize(dersig2)

        # s scalar must be an integer
        dersig2 = dersig[: Rsize + 4] + b"\x00" + dersig[Rsize + 5 :]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # Negative numbers are not allowed for s
        dersig2 = dersig[: Rsize + 6] + b"\x80" + dersig[Rsize + 7 :]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # Invalid null bytes at the start of s
        dersig2 = dersig[: Rsize + 6] + b"\x00\x00" + dersig[Rsize + 8 :]
        self.assertRaises(ValueError, _deserialize, dersig2)

        # sighash size > 1
        self.assertRaises(ValueError, _serialize, *sig, sighash_all + b"\x01")

        # negative signature scalar
        sig2 = -1, sig[1]
        self.assertRaises(ValueError, _serialize, *sig2, sighash_all)

        # Invalid sighash type b'\x00'
        self.assertRaises(ValueError, _deserialize, dersig[:-1] + b"\x00")
        # _deserialize(dersig[:-1] + b'\x00')


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import json
import unittest
import hashlib
from os import path

from btclib import dsa
from btclib.curvemult import mult
from btclib.curves import CURVES
from btclib.rfc6979 import rfc6979


class Testrfc6979(unittest.TestCase):
    def test_rfc6979(self):
        # source: https://bitcointalk.org/index.php?topic=285142.40
        msg = "Satoshi Nakamoto"
        x = 0x1
        k = 0x8F8A276C19F4149656B280621E358CCE24F5F52542772691EE69063B74F15D15
        k2 = rfc6979(msg, x, hf=hashlib.sha256)
        self.assertEqual(k, k2)

    def test_rfc6979_example(self):
        class _helper:
            def __init__(self, n: int) -> None:
                self.n = n
                self.nlen = n.bit_length()
                self.nsize = (self.nlen + 7) // 8

        # source: https://tools.ietf.org/html/rfc6979 section A.1

        fake_ec = _helper(0x4000000000000000000020108A2E0CC0D99F8A5EF)
        x = 0x09A4D6792295A7F730FC3F2B49CBC0F62E862272F
        msg = "sample"
        k = 0x23AF4074C90A02B3FE61D286D5C87F425E6BDD81B
        self.assertEqual(k, rfc6979(msg, x, fake_ec))

    def test_rfc6979_tv(self):

        fname = "rfc6979.json"
        filename = path.join(path.dirname(__file__), "test_data", fname)
        with open(filename, "r") as f:
            test_dict = json.load(f)

        for ec_name in test_dict:
            ec = CURVES[ec_name]
            test_vectors = test_dict[ec_name]
            for x, x_U, y_U, hf, msg, k, r, s in test_vectors:
                x = int(x, 16)
                # test RFC6979 implementation
                k2 = rfc6979(msg, x, ec, eval("hashlib." + hf))
                self.assertEqual(k, hex(k2))
                # test RFC6979 usage in DSA
                sig = dsa.sign(msg, x, k2, ec, eval("hashlib." + hf))
                self.assertEqual(r, hex(sig[0]))
                self.assertIn(s, (hex(sig[1]), hex(ec.n - sig[1])))
                # test that RFC6979 is the default nonce for DSA
                sig = dsa.sign(msg, x, k=None, ec=ec, hf=eval("hashlib." + hf))
                self.assertEqual(r, hex(sig[0]))
                self.assertIn(s, (hex(sig[1]), hex(ec.n - sig[1])))
                # test key-pair coherence
                U = mult(x, ec.G, ec)
                self.assertEqual((int(x_U, 16), int(y_U, 16)), U)
                # test signature validity
                dsa._verify(msg, U, sig, ec, hf=eval("hashlib." + hf))


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""BIP32 Hierarchical Deterministic Wallet functions.

A deterministic wallet is a hash-chain of private/public key pairs that
derives from a single root, which is the only element requiring backup.
Moreover, there are schemes where public keys can be calculated without
accessing private keys.

A hierarchical deterministic wallet is a tree of multiple hash-chains,
derived from a single root, allowing for selective sharing of keypair
chains.

Here, the HD wallet is implemented according to BIP32 bitcoin standard
https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki.

A BIP32 extended key is 78 bytes:

- [  : 4] version
- [ 4: 5] depth in the derivation path
- [ 5: 9] parent fingerprint
- [ 9:13] index
- [13:45] chain code
- [45:78] compressed pubkey or [0x00][prvkey]
"""

import copy
import hmac
from typing import List, Optional, Tuple

from . import bip39, electrum
from .alias import INF, BIP32Key, BIP32KeyDict, Octets, Path, Point
from .base58 import b58decode, b58encode
from .curvemult import mult
from .curves import secp256k1 as ec
from .mnemonic import Mnemonic
from .network import (
    _NETWORKS,
    _P2WPKH_PRV_PREFIXES,
    _XPRV_PREFIXES,
    _XPRV_VERSIONS_ALL,
    _XPUB_VERSIONS_ALL,
    NETWORKS,
)
from .secpoint import bytes_from_point, point_from_octets
from .utils import bytes_from_octets, hash160


class ExtendedBIP32KeyDict(BIP32KeyDict):
    # extensions used to cache intemediate results
    # in multi-level derivation: do not rely on them elsewhere
    q: int  # non-zero for private key only
    Q: Point  # non-Infinity for public key only


def _check_version_key(version: bytes, key: bytes) -> None:

    if version in _XPRV_VERSIONS_ALL:
        if key[0] != 0:
            raise ValueError("prv_version/pubkey mismatch")
    elif version in _XPUB_VERSIONS_ALL:
        if key[0] not in (2, 3):
            raise ValueError("pub_version/prvkey mismatch")
    else:
        raise ValueError(f"unknown extended key version {version.hex()}")


def _check_depth_pfp_index(depth: int, pfp: bytes, i: bytes) -> None:

    if depth < 0 or depth > 255:
        raise ValueError(f"Invalid BIP32 depth ({depth})")
    elif depth == 0:
        if pfp != b"\x00\x00\x00\x00":
            msg = f"Zero depth with non-zero parent_fingerprint ({pfp.hex()})"
            raise ValueError(msg)
        if i != b"\x00\x00\x00\x00":
            msg = f"Zero depth with non-zero index {i.hex()}"
            raise ValueError(msg)
    else:
        if pfp == b"\x00\x00\x00\x00":
            msg = f"Non-zero depth ({depth}) "
            msg += f"with zero parent_fingerprint ({pfp.hex()})"
            raise ValueError(msg)


def deserialize(xkey: BIP32Key) -> ExtendedBIP32KeyDict:

    d: ExtendedBIP32KeyDict
    if isinstance(xkey, dict):
        # no idea why mypy does complain about the following cleaner line
        # d = {**xkey, "q": 0, "Q": INF}
        # so, while waiting for the even better python 3.9
        # d = xkey | {"q": 0, "Q": INF}
        # let's make mypy happy with boring code like the following
        d = {
            "version": xkey["version"],
            "depth": xkey["depth"],
            "parent_fingerprint": xkey["parent_fingerprint"],
            "index": xkey["index"],
            "chain_code": xkey["chain_code"],
            "key": xkey["key"],
            # extensions
            "q": 0,  # non zero only if xprv
            "Q": INF,  # non INF only if xpub
        }
    else:
        if isinstance(xkey, str):
            xkey = xkey.strip()
        xkey = b58decode(xkey, 78)
        d = {
            "version": xkey[:4],
            "depth": xkey[4],
            "parent_fingerprint": xkey[5:9],
            "index": xkey[9:13],
            "chain_code": xkey[13:45],
            "key": xkey[45:],
            # extensions
            "q": 0,  # non zero only if xprv
            "Q": INF,  # non INF only if xpub
        }

    _check_version_key(d["version"], d["key"])
    _check_depth_pfp_index(d["depth"], d["parent_fingerprint"], d["index"])

    # calculate d["q"] and d["Q"]
    if d["key"][0] == 0:
        q = int.from_bytes(d["key"][1:], byteorder="big")
        if not 0 < q < ec.n:
            raise ValueError(f"Private key {hex(q).upper()} not in [1, n-1]")
        d["q"] = q
        d["Q"] = INF
    else:  # must be public (already checked by _check_version_key)
        d["q"] = 0
        d["Q"] = point_from_octets(d["key"], ec)

    return d


def serialize(d: BIP32KeyDict) -> bytes:

    if len(d["key"]) != 33:
        m = f"Invalid {len(d['key'])}-bytes BIP32 key length"
        raise ValueError(m)
    # version length is checked in _check_version_key
    _check_version_key(d["version"], d["key"])
    t = d["version"]

    if len(d["parent_fingerprint"]) != 4:
        m = f"Invalid {len(d['parent_fingerprint'])}-bytes "
        m += "BIP32 parent_fingerprint length"
        raise ValueError(m)
    if len(d["index"]) != 4:
        m = f"Invalid {len(d['index'])}-bytes BIP32 index length"
        raise ValueError(m)
    _check_depth_pfp_index(d["depth"], d["parent_fingerprint"], d["index"])
    t += d["depth"].to_bytes(1, "big")
    t += d["parent_fingerprint"]
    t += d["index"]

    if len(d["chain_code"]) != 32:
        m = f"Invalid {len(d['chain_code'])}-bytes BIP32 chain_code length"
        raise ValueError(m)
    t += d["chain_code"]

    # already checked in _check_version_key
    t += d["key"]

    return b58encode(t, 78)


def rootxprv_from_seed(seed: Octets, version: Optional[Octets] = None) -> bytes:
    """Return BIP32 root master extended private key from seed."""

    seed = bytes_from_octets(seed)
    hd = hmac.digest(b"Bitcoin seed", seed, "sha512")
    k = b"\x00" + hd[:32]
    if version is None:
        v = NETWORKS["mainnet"]["bip32_prv"]
    else:
        v = bytes_from_octets(version)
    if v not in _XPRV_VERSIONS_ALL:
        raise ValueError(f"unknown extended private key version {v!r}")

    d: BIP32KeyDict = {
        "version": v,
        "depth": 0,
        "parent_fingerprint": b"\x00\x00\x00\x00",
        "index": b"\x00\x00\x00\x00",
        "chain_code": hd[32:],
        "key": k,
    }
    return serialize(d)


def mxprv_from_bip39_mnemonic(
    mnemonic: Mnemonic, passphrase: str = "", version: Optional[Octets] = None
) -> bytes:
    """Return BIP32 root master extended private key from BIP39 mnemonic."""

    seed = bip39.seed_from_mnemonic(mnemonic, passphrase)
    return rootxprv_from_seed(seed, version)


def mxprv_from_electrum_mnemonic(
    mnemonic: Mnemonic, passphrase: str = "", network: str = "mainnet"
) -> bytes:
    """Return BIP32 master extended private key from Electrum mnemonic.

    Note that for a "standard" mnemonic the derivation path is "m",
    for a "segwit" mnemonic it is "m/0h" instead.
    """

    version, seed = electrum._seed_from_mnemonic(mnemonic, passphrase)
    network_index = _NETWORKS.index(network)

    if version == "standard":
        xversion = _XPRV_PREFIXES[network_index]
        return rootxprv_from_seed(seed, xversion)
    elif version == "segwit":
        xversion = _P2WPKH_PRV_PREFIXES[network_index]
        rootxprv = rootxprv_from_seed(seed, xversion)
        return derive(rootxprv, 0x80000000)  # "m/0h"
    else:
        raise ValueError(f"Unmanaged electrum mnemonic version ({version})")


def xpub_from_xprv(xprv: BIP32Key) -> bytes:
    """Neutered Derivation (ND).

    Derivation of the extended public key corresponding to an extended
    private key (“neutered” as it removes the ability to sign transactions).
    """

    if isinstance(xprv, dict):
        xprv = copy.copy(xprv)
    else:
        xprv = deserialize(xprv)

    if xprv["key"][0] != 0:
        raise ValueError(f"Not a private key: {serialize(xprv).decode()}")

    i = _XPRV_VERSIONS_ALL.index(xprv["version"])
    xprv["version"] = _XPUB_VERSIONS_ALL[i]

    q = int.from_bytes(xprv["key"][1:], byteorder="big")
    Q = mult(q)
    xprv["key"] = bytes_from_point(Q)

    return serialize(xprv)


def _ckd(d: ExtendedBIP32KeyDict, index: bytes) -> None:

    # d is a prvkey
    if d["key"][0] == 0:
        d["depth"] += 1
        Pbytes = bytes_from_point(mult(d["q"]))
        d["parent_fingerprint"] = hash160(Pbytes)[:4]
        d["index"] = index
        if index[0] >= 0x80:  # hardened derivation
            h = hmac.digest(d["chain_code"], d["key"] + index, "sha512")
        else:  # normal derivation
            h = hmac.digest(d["chain_code"], Pbytes + index, "sha512")
        d["chain_code"] = h[32:]
        offset = int.from_bytes(h[:32], byteorder="big")
        d["q"] = (d["q"] + offset) % ec.n
        d["key"] = b"\x00" + d["q"].to_bytes(32, "big")
        d["Q"] = INF
    # d is a pubkey
    else:
        if index[0] >= 0x80:
            raise ValueError("hardened derivation from pubkey is impossible")
        d["depth"] += 1
        d["parent_fingerprint"] = hash160(d["key"])[:4]
        d["index"] = index
        h = hmac.digest(d["chain_code"], d["key"] + index, "sha512")
        d["chain_code"] = h[32:]
        offset = int.from_bytes(h[:32], byteorder="big")
        Offset = mult(offset)
        d["Q"] = ec.add(d["Q"], Offset)
        d["key"] = bytes_from_point(d["Q"])
        d["q"] = 0


def _indexes_from_path(path: str) -> Tuple[List[bytes], bool]:

    steps = [x.strip() for x in path.split("/")]
    if steps[0] in ("m", "M"):
        absolute = True
    elif steps[0] == ".":
        absolute = False
    elif steps[0] == "":
        raise ValueError("Empty derivation path")
    else:
        raise ValueError(f"Invalid derivation path root: {steps[0]}")

    indexes: List[bytes] = list()
    for step in steps[1:]:
        if step == "":  # extra slash
            continue
        elif step[-1] in ("'", "H", "h"):
            index = int(step[:-1]) + 0x80000000
        else:
            index = int(step)
        indexes.append(index.to_bytes(4, "big"))

    if len(indexes) > 255:
        raise ValueError(f"Derivation path depth {len(indexes)}>255")
    return indexes, absolute


def derive(xkey: BIP32Key, path: Path) -> bytes:
    """Derive an extended key across a path spanning multiple depth levels.

    Derivation is according to:

    - absolute path as "m/44h/0'/1H/0/10" string
    - relative path as "./0/10" string
    - relative path as iterable integer indexes
    - relative one level child derivation with single integer index
    - relative one level child derivation with single 4-bytes index

    Path is case/blank/extra-slash insensitive
    (e.g. "M /44h / 0' /1H // 0/ 10 / ").
    """

    xkey = deserialize(xkey)

    if isinstance(path, str):
        path = path.strip()
        indexes, absolute = _indexes_from_path(path)
        if absolute and xkey["depth"] != 0:
            msg = "Absolute derivation path for non-root master key"
            raise ValueError(msg)
    elif isinstance(path, int):
        indexes = [path.to_bytes(4, byteorder="big")]
    elif isinstance(path, bytes):
        if len(path) != 4:
            raise ValueError(f"Index must be 4-bytes, not {len(path)}")
        indexes = [path]
    else:
        indexes = [i.to_bytes(4, byteorder="big") for i in path]

    final_depth = xkey["depth"] + len(indexes)
    if final_depth > 255:
        raise ValueError(f"Derivation path final depth {final_depth}>255")

    for index in indexes:
        _ckd(xkey, index)

    return serialize(xkey)


def crack_prvkey(parent_xpub: BIP32Key, child_xprv: BIP32Key) -> bytes:

    if isinstance(parent_xpub, dict):
        p = copy.copy(parent_xpub)
    else:
        p = deserialize(parent_xpub)

    if p["key"][0] not in (2, 3):
        m = "Extended parent key is not a public key: "
        m += f"{serialize(p).decode()}"
        raise ValueError(m)

    if isinstance(child_xprv, dict):
        c = child_xprv
    else:
        c = deserialize(child_xprv)
    if c["key"][0] != 0:
        m = f"Extended child key is not a private key: "
        m += f"{serialize(c).decode()}"
        raise ValueError(m)

    # check depth
    if c["depth"] != p["depth"] + 1:
        raise ValueError("not a parent's child: wrong depth relation")

    # check fingerprint
    if c["parent_fingerprint"] != hash160(p["key"])[:4]:
        raise ValueError("not a parent's child: wrong parent fingerprint")

    # check normal derivation
    if c["index"][0] >= 0x80:
        raise ValueError("hardened child derivation")

    p["version"] = c["version"]

    h = hmac.digest(p["chain_code"], p["key"] + c["index"], "sha512")
    child_q = int.from_bytes(c["key"][1:], byteorder="big")
    offset = int.from_bytes(h[:32], byteorder="big")
    parent_q = (child_q - offset) % ec.n
    p["key"] = b"\x00" + parent_q.to_bytes(32, byteorder="big")

    return serialize(p)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import secrets
import unittest
from typing import List

from btclib.alias import INF, INFJ
from btclib.curve import _mult_aff, _mult_jac
from btclib.curvemult import double_mult, mult, multi_mult
from btclib.curves import secp256k1
from btclib.tests.test_curves import low_card_curves

ec23_31 = low_card_curves["ec23_31"]


class TestEllipticCurve(unittest.TestCase):
    def test_mult(self):
        for ec in low_card_curves.values():
            for q in range(ec.n):
                Q = _mult_aff(q, ec.G, ec)
                QJ = _mult_jac(q, ec.GJ, ec)
                Q2 = ec._aff_from_jac(QJ)
                self.assertEqual(Q, Q2)
        # with last curve
        self.assertEqual(INF, _mult_aff(3, INF, ec))
        self.assertEqual(INFJ, _mult_jac(3, INFJ, ec))

    def test_shamir(self):
        ec = ec23_31
        for k1 in range(ec.n):
            for k2 in range(ec.n):
                shamir = double_mult(k1, ec.G, k2, ec.G, ec)
                std = ec.add(mult(k1, ec.G, ec), mult(k2, ec.G, ec))
                self.assertEqual(shamir, std)
                shamir = double_mult(k1, INF, k2, ec.G, ec)
                std = ec.add(mult(k1, INF, ec), mult(k2, ec.G, ec))
                self.assertEqual(shamir, std)
                shamir = double_mult(k1, ec.G, k2, INF, ec)
                std = ec.add(mult(k1, ec.G, ec), mult(k2, INF, ec))
                self.assertEqual(shamir, std)

    def test_boscoster(self):
        ec = secp256k1

        k: List[int] = list()
        ksum = 0
        for i in range(11):
            k.append(secrets.randbits(ec.nlen) % ec.n)
            ksum += k[i]

        P = [ec.G] * len(k)
        boscoster = multi_mult(k, P, ec)
        self.assertEqual(boscoster, mult(ksum, ec.G, ec))

        # mismatch between scalar length and Points length
        P = [ec.G] * (len(k) - 1)
        self.assertRaises(ValueError, multi_mult, k, P, ec)
        # multi_mult(k, P, ec)


def test_double_mult():
    H = (
        0x50929B74C1A04954B78B4B6035E97A5E078A5A0F28EC96D547BFEE9ACE803AC0,
        0x31D3C6863973926E049E637CB1B5F40A36DAC28AF1766968C30C2313F3A38904,
    )
    G = secp256k1.G

    # 0*G + 1*H
    T = double_mult(1, H, 0, G)
    assert T == H

    # 0*G + 2*H
    T = double_mult(2, H, 0, G)
    assert T == mult(2, H)

    # 0*G + 3*H
    T = double_mult(3, H, 0, G)
    assert T == mult(3, H)

    # 1*G + 0*H
    T = double_mult(0, H, 1, G)
    assert T == G

    # 2*G + 0*H
    T = double_mult(0, H, 2, G)
    assert T == mult(2, G)

    # 3*G + 0*H
    T = double_mult(0, H, 3, G)
    assert T == mult(3, G)

    # 0*G + 5*H
    T = double_mult(5, H, 0, G)
    assert T == mult(5, H)

    # 0*G - 5*H
    T = double_mult(-5, H, 0, G)
    assert T == mult(-5, H)

    # 1*G - 5*H
    U = double_mult(-5, H, 1, G)
    assert U == secp256k1.add(G, T)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest
from os import path

from btclib.mnemonic import _indexes_from_mnemonic, _mnemonic_from_indexes, _wordlists


class TestMnemonic(unittest.TestCase):
    def test_mnemonic(self):
        lang = "en"

        test_mnemonic = (
            "ozone drill grab fiber curtain grace "
            "pudding thank cruise elder eight picnic"
        )
        test_indexes = [1268, 535, 810, 685, 433, 811, 1385, 1790, 421, 570, 567, 1313]
        indexes = _indexes_from_mnemonic(test_mnemonic, lang)
        self.assertEqual(indexes, test_indexes)

        mnemonic = _mnemonic_from_indexes(test_indexes, lang)
        self.assertEqual(mnemonic, test_mnemonic)

    def test_wordlist_1(self):
        lang = "en"

        d = _wordlists.wordlist(lang)
        self.assertIsInstance(d, list)
        self.assertEqual(len(d), 2048)

        length = _wordlists.language_length(lang)
        self.assertEqual(length, 2048)

    def test_wordlist_2(self):
        lang = "fakeen"
        # missing file for language 'fakeen''
        self.assertRaises(ValueError, _wordlists.load_lang, lang)
        # _wordlists.load_lang(lang)

        # dictionary length (must be a power of two
        fname = "fakeenglish.txt"
        filename = path.join(path.dirname(__file__), "test_data", fname)
        self.assertRaises(ValueError, _wordlists.load_lang, lang, filename)
        # _wordlists.load_lang(lang, filename)

        # dinamically add a new language
        lang = "en2"
        fname = "english.txt"
        filename = path.join(path.dirname(__file__), "test_data", fname)
        _wordlists.load_lang(lang, filename)
        length = _wordlists.language_length(lang)
        self.assertEqual(length, 2048)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Network constants and associated functions."""

import copy
from collections import defaultdict
from typing import Dict, List, TypedDict, Union

from .curve import Curve
from .curves import secp256k1


class Network(TypedDict):
    curve: Curve
    wif: bytes
    p2pkh: bytes
    p2sh: bytes
    p2w: str
    bip32_prv: bytes
    bip32_pub: bytes
    slip32_p2wpkh_prv: bytes
    slip32_p2wpkh_pub: bytes
    slip32_p2wpkh_p2sh_prv: bytes
    slip32_p2wpkh_p2sh_pub: bytes
    slip32_p2wsh_prv: bytes
    slip32_p2wsh_pub: bytes
    slip32_p2wsh_p2sh_prv: bytes
    slip32_p2wsh_p2sh_pub: bytes


NETWORKS: Dict[str, Network] = defaultdict()

NETWORKS["mainnet"] = {
    "curve": secp256k1,
    "wif": b"\x80",  # base58wif starts with 'K' or 'L' if compressed else '5'
    "p2pkh": b"\x00",  # base58address starts with '1'
    "p2sh": b"\x05",  # base58address starts with '3'
    "p2w": "bc",  # bech32address starts with 'bc1'
    # slip32 "m / 44h / 0h" p2pkh or p2sh
    "bip32_prv": b"\x04\x88\xAD\xE4",  # xprv
    "bip32_pub": b"\x04\x88\xB2\x1E",  # xpub
    # slip32 "m / 49h / 0h" p2wpkh-p2sh (p2sh-wrapped-segwit)
    "slip32_p2wsh_p2sh_prv": b"\x04\x9D\x78\x78",  # yprv
    "slip32_p2wsh_p2sh_pub": b"\x04\x9D\x7C\xB2",  # ypub
    # slip32 p2wsh-p2sh (p2sh-wrapped-segwit)
    "slip32_p2wpkh_p2sh_prv": b"\x02\x95\xB0\x05",  # Yprv
    "slip32_p2wpkh_p2sh_pub": b"\x02\x95\xB4\x3F",  # Ypub
    # slip32 "m / 84h / 0h" p2wpkh (native-segwit)
    "slip32_p2wpkh_prv": b"\x04\xB2\x43\x0C",  # zprv
    "slip32_p2wpkh_pub": b"\x04\xB2\x47\x46",  # zpub
    # slip32 p2wsh (native-segwit)
    "slip32_p2wsh_prv": b"\x02\xAA\x7A\x99",  # Zprv
    "slip32_p2wsh_pub": b"\x02\xAA\x7E\xD3",  # Zpub
}

NETWORKS["testnet"] = {
    "curve": secp256k1,
    "wif": b"\xef",  # base58wif starts with 'c' if compressed else '9'
    "p2pkh": b"\x6f",  # base58address starts with 'm' or 'n'
    "p2sh": b"\xc4",  # base58address starts with '2'
    "p2w": "tb",  # bech32address starts with 'tb1'
    # slip32 "m / 44h / 1h" p2pkh or p2sh
    "bip32_prv": b"\x04\x35\x83\x94",  # tprv
    "bip32_pub": b"\x04\x35\x87\xCF",  # tpub
    # slip32 "m / 49h / 1h" p2wpkh-p2sh (p2sh-wrapped-segwit)
    "slip32_p2wsh_p2sh_prv": b"\x04\x4A\x4E\x28",  # uprv
    "slip32_p2wsh_p2sh_pub": b"\x04\x4A\x52\x62",  # upub
    # slip32 p2wsh-p2sh (p2sh-wrapped-segwit)
    "slip32_p2wpkh_p2sh_prv": b"\x02\x42\x85\xB5",  # Uprv
    "slip32_p2wpkh_p2sh_pub": b"\x02\x42\x89\xEF",  # Upub
    # slip32 "m / 84h / 1h" p2wpkh (native-segwit)
    "slip32_p2wpkh_prv": b"\x04\x5F\x18\xBC",  # vprv
    "slip32_p2wpkh_pub": b"\x04\x5F\x1C\xF6",  # vpub
    # slip32 p2wsh (native-segwit)
    "slip32_p2wsh_prv": b"\x02\x57\x50\x48",  # Vprv
    "slip32_p2wsh_pub": b"\x02\x57\x54\x83",  # Vpub
}

NETWORKS["regtest"] = copy.copy(NETWORKS["testnet"])
NETWORKS["regtest"]["p2w"] = "bcrt"  # bech32address starts with 'bcrt1'


def network_from_key_value(key: str, prefix: Union[str, bytes, Curve]) -> str:
    """Return network string from (key, value) pair.

    Warning: when used on 'regtest' it mostly returns 'testnet',
    which is not a problem as long as it is used for
    WIF/Base58Address/BIP32xkey
    because the two networks share the same prefixes.
    """
    for net in NETWORKS:
        if NETWORKS[net][key] == prefix:
            return net
    raise ValueError(f"No network has {key} = {prefix!r}")


_NETWORKS = [net for net in NETWORKS]
_P2PKH_PREFIXES = [NETWORKS[net]["p2pkh"] for net in NETWORKS]
_P2SH_PREFIXES = [NETWORKS[net]["p2sh"] for net in NETWORKS]

_XPRV_PREFIXES = [NETWORKS[net]["bip32_prv"] for net in NETWORKS]
_XPUB_PREFIXES = [NETWORKS[net]["bip32_pub"] for net in NETWORKS]

_P2WPKH_P2SH_PUB_PREFIXES = [NETWORKS[net]["slip32_p2wsh_p2sh_pub"] for net in NETWORKS]

_P2WPKH_PRV_PREFIXES = [NETWORKS[net]["slip32_p2wpkh_prv"] for net in NETWORKS]
_P2WPKH_PUB_PREFIXES = [NETWORKS[net]["slip32_p2wpkh_pub"] for net in NETWORKS]


def xpubversions_from_network(network: str) -> List[bytes]:
    network = network.lower()
    result = [
        NETWORKS[network]["bip32_pub"],
        NETWORKS[network]["slip32_p2wsh_p2sh_pub"],
        NETWORKS[network]["slip32_p2wpkh_p2sh_pub"],
        NETWORKS[network]["slip32_p2wpkh_pub"],
        NETWORKS[network]["slip32_p2wsh_pub"],
    ]
    return result


def xprvversions_from_network(network: str) -> List[bytes]:
    network = network.lower()
    result = [
        NETWORKS[network]["bip32_prv"],
        NETWORKS[network]["slip32_p2wsh_p2sh_prv"],
        NETWORKS[network]["slip32_p2wpkh_p2sh_prv"],
        NETWORKS[network]["slip32_p2wpkh_prv"],
        NETWORKS[network]["slip32_p2wsh_prv"],
    ]
    return result


# the following provides false match for regtest
# not a problem as long as it is used for WIF/Base58Address/BIP32xkey
# where the two networks share same prefixes.
_XPRV_VERSIONS_MAIN = [
    NETWORKS["mainnet"]["bip32_prv"],
    NETWORKS["mainnet"]["slip32_p2wsh_p2sh_prv"],
    NETWORKS["mainnet"]["slip32_p2wpkh_p2sh_prv"],
    NETWORKS["mainnet"]["slip32_p2wpkh_prv"],
    NETWORKS["mainnet"]["slip32_p2wsh_prv"],
]
_XPRV_VERSIONS_TEST = [
    NETWORKS["testnet"]["bip32_prv"],
    NETWORKS["testnet"]["slip32_p2wsh_p2sh_prv"],
    NETWORKS["testnet"]["slip32_p2wpkh_p2sh_prv"],
    NETWORKS["testnet"]["slip32_p2wpkh_prv"],
    NETWORKS["testnet"]["slip32_p2wsh_prv"],
]
_XPUB_VERSIONS_MAIN = [
    NETWORKS["mainnet"]["bip32_pub"],
    NETWORKS["mainnet"]["slip32_p2wsh_p2sh_pub"],
    NETWORKS["mainnet"]["slip32_p2wpkh_p2sh_pub"],
    NETWORKS["mainnet"]["slip32_p2wpkh_pub"],
    NETWORKS["mainnet"]["slip32_p2wsh_pub"],
]
_XPUB_VERSIONS_TEST = [
    NETWORKS["testnet"]["bip32_pub"],
    NETWORKS["testnet"]["slip32_p2wsh_p2sh_pub"],
    NETWORKS["testnet"]["slip32_p2wpkh_p2sh_pub"],
    NETWORKS["testnet"]["slip32_p2wpkh_pub"],
    NETWORKS["testnet"]["slip32_p2wsh_pub"],
]
_XPRV_VERSIONS_ALL = _XPRV_VERSIONS_MAIN + _XPRV_VERSIONS_TEST + _XPRV_VERSIONS_TEST
_XPUB_VERSIONS_ALL = _XPUB_VERSIONS_MAIN + _XPUB_VERSIONS_TEST + _XPUB_VERSIONS_TEST
_REPEATED_NETWORKS = [_NETWORKS[0]] * 5 + [_NETWORKS[1]] * 5 + [_NETWORKS[2]] * 5


def network_from_xkeyversion(xprvversion: bytes) -> str:
    """Return network string from the xkey version prefix.

    Warning: when used on 'regtest' it returns 'testnet', which is not
    a problem as long as it is used for WIF/Base58Address/BIP32Key
    because the two networks share the same prefixes.
    """
    try:
        index = _XPRV_VERSIONS_ALL.index(xprvversion)
    except Exception:
        index = _XPUB_VERSIONS_ALL.index(xprvversion)

    return _REPEATED_NETWORKS[index]


_CURVES = [NETWORKS[net]["curve"] for net in NETWORKS]


def curve_from_xpubversion(xpubversion: bytes) -> Curve:
    index = _XPUB_VERSIONS_ALL.index(xpubversion)
    return _CURVES[index]



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import csv
import secrets
import unittest
from hashlib import sha256 as hf
from os import path
from typing import List

from btclib import ssa
from btclib.alias import INF, Point
from btclib.curvemult import double_mult, mult
from btclib.curves import CURVES
from btclib.curves import secp256k1 as ec
from btclib.numbertheory import mod_inv
from btclib.pedersen import second_generator
from btclib.tests.test_curves import low_card_curves
from btclib.utils import int_from_bits

secp224k1 = CURVES['secp224k1']


class TestSSA(unittest.TestCase):

    def test_signature(self):
        """Basic tests"""

        q, x_Q = ssa.gen_keys()
        mhd = hf(b'Satoshi Nakamoto').digest()
        sig = ssa.sign(mhd, q, None)
        self.assertEqual(sig, ssa.deserialize(sig))
        ssa._verify(mhd, x_Q, sig, ec, hf)
        self.assertTrue(ssa.verify(mhd, x_Q, sig))

        fmhd = hf(b'Craig Wright').digest()
        self.assertRaises(AssertionError, ssa._verify, fmhd, x_Q, sig, ec, hf)

        fssasig = (sig[0], sig[1], sig[1])
        self.assertRaises(ValueError, ssa._verify, mhd, x_Q, fssasig, ec, hf)

        # y(sG - eP) is not a quadratic residue
        _, fQ = ssa.gen_keys(0x2)
        self.assertRaises(AssertionError, ssa._verify, mhd, fQ, sig, ec, hf)

        _, fQ = ssa.gen_keys(0x4)
        self.assertRaises(AssertionError, ssa._verify, mhd, fQ, sig, ec, hf)

        # not ec.pIsThreeModFour
        self.assertRaises(ValueError, ssa._verify,
                          mhd, x_Q, sig, secp224k1, hf)

        # verify: message of wrong size
        wrongmhd = mhd[:-1]
        self.assertRaises(ValueError, ssa._verify, wrongmhd, x_Q, sig, ec, hf)
        # ssa._verify(wrongmhd, x_Q, sig)

        # sign: message of wrong size
        self.assertRaises(ValueError, ssa.sign, wrongmhd, q, None)
        # ssa.sign(wrongmhd, q, None)

        # invalid (zero) challenge e
        self.assertRaises(ValueError, ssa._recover_pubkeys,
                          0, sig[0], sig[1], ec)
        # ssa._recover_pubkeys(0, sig)

        # not a BIP340 public key
        self.assertRaises(ValueError, ssa._to_bip340_point,
                          ["not", "a BIP340", "public key"])
        # ssa._to_bip340_point(["not", "a BIP340", "public key"])

    def test_bip340_vectors(self):
        """BIP340 (Schnorr) test vectors

        https://github.com/bitcoin/bips/blob/master/bip-0340/test-vectors.csv
        """
        fname = "bip340_test_vectors.csv"
        filename = path.join(path.dirname(__file__), "test_data", fname)
        with open(filename, newline='') as csvfile:
            reader = csv.reader(csvfile)
            # skip column headers while checking that there are 7 columns
            _, _, _, _, _, _, _ = reader.__next__()
            for row in reader:
                (index, seckey, pubkey, mhd, sig, result, comment) = row
                errmsg = f"Test vector #{int(index)}"
                if seckey != '':
                    seckey = bytes.fromhex(seckey)
                    _, pubkey_actual = ssa.gen_keys(seckey)
                    self.assertEqual(pubkey,
                                     hex(pubkey_actual).upper()[2:], errmsg)

                    sig_actual = ssa.serialize(*ssa.sign(mhd, seckey))
                    self.assertEqual(sig, sig_actual.hex().upper(), errmsg)

                result = result == 'TRUE'
                if comment:
                    errmsg += ": " + comment
                result_actual = ssa.verify(mhd, pubkey, sig)
                self.assertEqual(result, result_actual, errmsg)

    def test_low_cardinality(self):
        """test low-cardinality curves for all msg/key pairs."""
        # e_c.n has to be prime to sign
        prime = [11, 13, 17, 19]

        # for dsa it is possible to directy iterate on all
        # possible values for e;
        # for ssa we have to iterate over all possible hash values
        hsize = hf().digest_size
        H = [i.to_bytes(hsize, 'big') for i in range(max(prime) * 4)]
        # only low card curves or it would take forever
        for e_c in low_card_curves.values():
            if e_c.p in prime:  # only few curves or it would take too long
                # BIP340 Schnorr only applies to curve whose prime p = 3 %4
                if not e_c.pIsThreeModFour:
                    self.assertRaises(ValueError, ssa.sign, H[0], 1, None, e_c)
                    continue
                for q in range(1, e_c.n):  # all possible private keys
                    Q = mult(q, e_c.G, e_c)  # public key
                    if not e_c.has_square_y(Q):
                        q = e_c.n - q
                    for h in H:  # all possible hashed messages
                        k = ssa.k(h, q, e_c, hf)
                        K = mult(k, e_c.G, e_c)
                        if not e_c.has_square_y(K):
                            k = e_c.n - k
                        x_K = K[0]
                        try:
                            c = ssa._challenge(x_K, Q[0], h, e_c, hf)
                        except Exception:
                            pass
                        else:
                            s = (k + c * q) % e_c.n
                            sig = ssa.sign(h, q, None, e_c)
                            self.assertEqual((x_K, s), sig)
                            # valid signature must validate
                            self.assertIsNone(ssa._verify(h, Q, sig, e_c, hf))

                            if c != 0:  # FIXME
                                x_Q = ssa._recover_pubkeys(c, x_K, s, e_c)
                                self.assertEqual(Q[0], x_Q)

    def test_batch_validation(self):
        hsize = hf().digest_size
        hlen = hsize * 8

        ms = []
        Qs = []
        sigs = []
        ms.append(secrets.randbits(hlen).to_bytes(hsize, 'big'))
        q = 1 + secrets.randbelow(ec.n - 1)
        # bytes version
        Qs.append(mult(q, ec.G, ec)[0].to_bytes(ec.psize, 'big'))
        sigs.append(ssa.sign(ms[0], q, None, ec, hf))
        # test with only 1 sig
        ssa._batch_verify(ms, Qs, sigs, ec, hf)
        for _ in range(3):
            mhd = secrets.randbits(hlen).to_bytes(hsize, 'big')
            ms.append(mhd)
            q = 1 + secrets.randbelow(ec.n - 1)
            # Point version
            Qs.append(mult(q, ec.G, ec))
            sigs.append(ssa.sign(mhd, q, None, ec, hf))
        ssa._batch_verify(ms, Qs, sigs, ec, hf)
        self.assertTrue(ssa.batch_verify(ms, Qs, sigs, ec, hf))

        # invalid sig
        ms.append(ms[0])
        sigs.append(sigs[1])
        Qs.append(Qs[0])
        self.assertFalse(ssa.batch_verify(ms, Qs, sigs, ec, hf))
        self.assertRaises(AssertionError, ssa._batch_verify,
                          ms, Qs, sigs, ec, hf)
        # ssa._batch_verify(ms, Qs, sigs, ec, hf)
        sigs[-1] = sigs[0]  # valid again

        # Invalid size: 31 bytes instead of 32
        ms[-1] = ms[0][:-1]
        self.assertRaises(ValueError, ssa._batch_verify, ms, Qs, sigs, ec, hf)
        # ssa._batch_verify(ms, Qs, sigs, ec, hf)
        ms[-1] = ms[0]  # valid again

        # mismatch between number of pubkeys (5) and number of messages (6)
        ms.append(ms[0])  # add extra message
        self.assertRaises(ValueError, ssa._batch_verify, ms, Qs, sigs, ec, hf)
        # ssa._batch_verify(ms, Qs, sigs, ec, hf)
        ms.pop()  # valid again

        # mismatch between number of pubkeys (5) and number of signatures (6)
        sigs.append(sigs[0])  # add extra sig
        self.assertRaises(ValueError, ssa._batch_verify, ms, Qs, sigs, ec, hf)
        # ssa._batch_verify(ms, Qs, sigs, ec, hf)
        sigs.pop()  # valid again

        # field prime p is not equal to 3 (mod 4)
        self.assertRaises(ValueError, ssa._batch_verify,
                          ms, Qs, sigs, secp224k1, hf)
        # ssa._batch_verify(ms, Qs, sigs, secp224k1, hf)

    def test_threshold(self):
        """testing 2-of-3 threshold signature (Pedersen secret sharing)"""

        # parameters
        m = 2
        H = second_generator(ec, hf)
        mhd = hf(b'message to sign').digest()

        # FIRST PHASE: key pair generation ###################################

        # 1.1 signer one acting as the dealer
        commits1: List[Point] = list()
        q1, _ = ssa.gen_keys()
        q1_prime, _ = ssa.gen_keys()
        commits1.append(double_mult(q1_prime, H, q1, ec.G))
        # sharing polynomials
        f1 = [q1]
        f1_prime = [q1_prime]
        for i in range(1, m):
            f1.append(ssa.gen_keys()[0])
            f1_prime.append(ssa.gen_keys()[0])
            commits1.append(double_mult(f1_prime[i], H, f1[i], ec.G))
        # shares of the secret
        alpha12 = 0  # share of q1 belonging to signer two
        alpha12_prime = 0
        alpha13 = 0  # share of q1 belonging to signer three
        alpha13_prime = 0
        for i in range(m):
            alpha12 += (f1[i] * pow(2, i)) % ec.n
            alpha12_prime += (f1_prime[i] * pow(2, i)) % ec.n
            alpha13 += (f1[i] * pow(3, i)) % ec.n
            alpha13_prime += (f1_prime[i] * pow(3, i)) % ec.n
        # signer two verifies consistency of his share
        RHS = INF
        for i in range(m):
            RHS = ec.add(RHS, mult(pow(2, i), commits1[i]))
        t = double_mult(alpha12_prime, H, alpha12, ec.G)
        assert t == RHS, 'signer one is cheating'
        # signer three verifies consistency of his share
        RHS = INF
        for i in range(m):
            RHS = ec.add(RHS, mult(pow(3, i), commits1[i]))
        t = double_mult(alpha13_prime, H, alpha13, ec.G)
        assert t == RHS, 'signer one is cheating'

        # 1.2 signer two acting as the dealer
        commits2: List[Point] = list()
        q2, _ = ssa.gen_keys()
        q2_prime, _ = ssa.gen_keys()
        commits2.append(double_mult(q2_prime, H, q2, ec.G))
        # sharing polynomials
        f2 = [q2]
        f2_prime = [q2_prime]
        for i in range(1, m):
            f2.append(ssa.gen_keys()[0])
            f2_prime.append(ssa.gen_keys()[0])
            commits2.append(double_mult(f2_prime[i], H, f2[i], ec.G))
        # shares of the secret
        alpha21 = 0  # share of q2 belonging to signer one
        alpha21_prime = 0
        alpha23 = 0  # share of q2 belonging to signer three
        alpha23_prime = 0
        for i in range(m):
            alpha21 += (f2[i] * pow(1, i)) % ec.n
            alpha21_prime += (f2_prime[i] * pow(1, i)) % ec.n
            alpha23 += (f2[i] * pow(3, i)) % ec.n
            alpha23_prime += (f2_prime[i] * pow(3, i)) % ec.n
        # signer one verifies consistency of his share
        RHS = INF
        for i in range(m):
            RHS = ec.add(RHS, mult(pow(1, i), commits2[i]))
        t = double_mult(alpha21_prime, H, alpha21, ec.G)
        assert t == RHS, 'signer two is cheating'
        # signer three verifies consistency of his share
        RHS = INF
        for i in range(m):
            RHS = ec.add(RHS, mult(pow(3, i), commits2[i]))
        t = double_mult(alpha23_prime, H, alpha23, ec.G)
        assert t == RHS, 'signer two is cheating'

        # 1.3 signer three acting as the dealer
        commits3: List[Point] = list()
        q3, _ = ssa.gen_keys()
        q3_prime, _ = ssa.gen_keys()
        commits3.append(double_mult(q3_prime, H, q3, ec.G))
        # sharing polynomials
        f3 = [q3]
        f3_prime = [q3_prime]
        for i in range(1, m):
            f3.append(ssa.gen_keys()[0])
            f3_prime.append(ssa.gen_keys()[0])
            commits3.append(double_mult(f3_prime[i], H, f3[i], ec.G))
        # shares of the secret
        alpha31 = 0  # share of q3 belonging to signer one
        alpha31_prime = 0
        alpha32 = 0  # share of q3 belonging to signer two
        alpha32_prime = 0
        for i in range(m):
            alpha31 += (f3[i] * pow(1, i)) % ec.n
            alpha31_prime += (f3_prime[i] * pow(1, i)) % ec.n
            alpha32 += (f3[i] * pow(2, i)) % ec.n
            alpha32_prime += (f3_prime[i] * pow(2, i)) % ec.n
        # signer one verifies consistency of his share
        RHS = INF
        for i in range(m):
            RHS = ec.add(RHS, mult(pow(1, i), commits3[i]))
        t = double_mult(alpha31_prime, H, alpha31, ec.G)
        assert t == RHS, 'signer three is cheating'
        # signer two verifies consistency of his share
        RHS = INF
        for i in range(m):
            RHS = ec.add(RHS, mult(pow(2, i), commits3[i]))
        t = double_mult(alpha32_prime, H, alpha32, ec.G)
        assert t == RHS, 'signer three is cheating'
        # shares of the secret key q = q1 + q2 + q3
        alpha1 = (alpha21 + alpha31) % ec.n
        alpha2 = (alpha12 + alpha32) % ec.n
        alpha3 = (alpha13 + alpha23) % ec.n
        for i in range(m):
            alpha1 += (f1[i] * pow(1, i)) % ec.n
            alpha2 += (f2[i] * pow(2, i)) % ec.n
            alpha3 += (f3[i] * pow(3, i)) % ec.n

        # 1.4 it's time to recover the public key
        # each participant i = 1, 2, 3 shares Qi as follows
        # Q = Q1 + Q2 + Q3 = (q1 + q2 + q3) G
        A1: List[Point] = list()
        A2: List[Point] = list()
        A3: List[Point] = list()
        for i in range(m):
            A1.append(mult(f1[i]))
            A2.append(mult(f2[i]))
            A3.append(mult(f3[i]))
        # signer one checks others' values
        RHS2 = INF
        RHS3 = INF
        for i in range(m):
            RHS2 = ec.add(RHS2, mult(pow(1, i), A2[i]))
            RHS3 = ec.add(RHS3, mult(pow(1, i), A3[i]))
        assert mult(alpha21) == RHS2, 'signer two is cheating'
        assert mult(alpha31) == RHS3, 'signer three is cheating'
        # signer two checks others' values
        RHS1 = INF
        RHS3 = INF
        for i in range(m):
            RHS1 = ec.add(RHS1, mult(pow(2, i), A1[i]))
            RHS3 = ec.add(RHS3, mult(pow(2, i), A3[i]))
        assert mult(alpha12) == RHS1, 'signer one is cheating'
        assert mult(alpha32) == RHS3, 'signer three is cheating'
        # signer three checks others' values
        RHS1 = INF
        RHS2 = INF
        for i in range(m):
            RHS1 = ec.add(RHS1, mult(pow(3, i), A1[i]))
            RHS2 = ec.add(RHS2, mult(pow(3, i), A2[i]))
        assert mult(alpha13) == RHS1, 'signer one is cheating'
        assert mult(alpha23) == RHS2, 'signer two is cheating'
        # commitment at the global sharing polynomial
        A: List[Point] = list()
        for i in range(m):
            A.append(ec.add(A1[i], ec.add(A2[i], A3[i])))

        # aggregated public key
        Q = A[0]
        if not ec.has_square_y(Q):
            # print('Q has been negated')
            A[1] = ec.negate(A[1])  # pragma: no cover
            alpha1 = ec.n - alpha1  # pragma: no cover
            alpha2 = ec.n - alpha2  # pragma: no cover
            alpha3 = ec.n - alpha3  # pragma: no cover
            Q = ec.negate(Q)  # pragma: no cover

        # SECOND PHASE: generation of the nonces' pair  ######################
        # Assume signer one and three want to sign

        # 2.1 signer one acting as the dealer
        commits1: List[Point] = list()
        k1 = ssa.k(mhd, q1, ec, hf)
        k1_prime = ssa.k(mhd, q1_prime, ec, hf)
        commits1.append(double_mult(k1_prime, H, k1, ec.G))
        # sharing polynomials
        f1 = [k1]
        f1_prime = [k1_prime]
        for i in range(1, m):
            f1.append(ssa.gen_keys()[0])
            f1_prime.append(ssa.gen_keys()[0])
            commits1.append(double_mult(f1_prime[i], H, f1[i], ec.G))
        # shares of the secret
        beta13 = 0  # share of k1 belonging to signer three
        beta13_prime = 0
        for i in range(m):
            beta13 += (f1[i] * pow(3, i)) % ec.n
            beta13_prime += (f1_prime[i] * pow(3, i)) % ec.n
        # signer three verifies consistency of his share
        RHS = INF
        for i in range(m):
            RHS = ec.add(RHS, mult(pow(3, i), commits1[i]))
        t = double_mult(beta13_prime, H, beta13, ec.G)
        assert t == RHS, 'signer one is cheating'

        # 2.2 signer three acting as the dealer
        commits3: List[Point] = list()
        k3 = ssa.k(mhd, q3, ec, hf)
        k3_prime = ssa.k(mhd, q3_prime, ec, hf)
        commits3.append(double_mult(k3_prime, H, k3, ec.G))
        # sharing polynomials
        f3 = [k3]
        f3_prime = [k3_prime]
        for i in range(1, m):
            f3.append(ssa.gen_keys()[0])
            f3_prime.append(ssa.gen_keys()[0])
            commits3.append(double_mult(f3_prime[i], H, f3[i], ec.G))
        # shares of the secret
        beta31 = 0  # share of k3 belonging to signer one
        beta31_prime = 0
        for i in range(m):
            beta31 += (f3[i] * pow(1, i)) % ec.n
            beta31_prime += (f3_prime[i] * pow(1, i)) % ec.n
        # signer one verifies consistency of his share
        RHS = INF
        for i in range(m):
            RHS = ec.add(RHS, mult(pow(1, i), commits3[i]))
        t = double_mult(beta31_prime, H, beta31, ec.G)
        assert t == RHS, 'signer three is cheating'

        # 2.3 shares of the secret nonce
        beta1 = beta31 % ec.n
        beta3 = beta13 % ec.n
        for i in range(m):
            beta1 += (f1[i] * pow(1, i)) % ec.n
            beta3 += (f3[i] * pow(3, i)) % ec.n

        # 2.4 it's time to recover the public nonce
        # each participant i = 1, 3 shares Qi as follows
        B1: List[Point] = list()
        B3: List[Point] = list()
        for i in range(m):
            B1.append(mult(f1[i]))
            B3.append(mult(f3[i]))

        # signer one checks values from signer three
        RHS3 = INF
        for i in range(m):
            RHS3 = ec.add(RHS3, mult(pow(1, i), B3[i]))
        assert mult(beta31) == RHS3, 'signer three is cheating'

        # signer three checks values from signer one
        RHS1 = INF
        for i in range(m):
            RHS1 = ec.add(RHS1, mult(pow(3, i), B1[i]))
        assert mult(beta13) == RHS1, 'signer one is cheating'

        # commitment at the global sharing polynomial
        B: List[Point] = list()
        for i in range(m):
            B.append(ec.add(B1[i], B3[i]))

        # aggregated public nonce
        K = B[0]
        if not ec.has_square_y(K):
            # print('K has been negated')
            B[1] = ec.negate(B[1])  # pragma: no cover
            beta1 = ec.n - beta1  # pragma: no cover
            beta3 = ec.n - beta3  # pragma: no cover
            K = ec.negate(K)  # pragma: no cover

        # PHASE THREE: signature generation ###

        # partial signatures
        e = ssa._challenge(K[0], Q[0], mhd, ec, hf)
        gamma1 = (beta1 + e * alpha1) % ec.n
        gamma3 = (beta3 + e * alpha3) % ec.n

        # each participant verifies the other partial signatures

        # signer one
        RHS3 = ec.add(K, mult(e, Q))
        for i in range(1, m):
            temp = double_mult(pow(3, i), B[i], e * pow(3, i), A[i])
            RHS3 = ec.add(RHS3, temp)
        assert mult(gamma3) == RHS3, 'signer three is cheating'

        # signer three
        RHS1 = ec.add(K, mult(e, Q))
        for i in range(1, m):
            temp = double_mult(pow(1, i), B[i], e * pow(1, i), A[i])
            RHS1 = ec.add(RHS1, temp)
        assert mult(gamma1) == RHS1, 'signer one is cheating'

        # PHASE FOUR: aggregating the signature ###
        omega1 = 3 * mod_inv(3 - 1, ec.n) % ec.n
        omega3 = 1 * mod_inv(1 - 3, ec.n) % ec.n
        sigma = (gamma1 * omega1 + gamma3 * omega3) % ec.n

        sig = K[0], sigma

        self.assertTrue(ssa.verify(mhd, Q, sig))

        # ADDITIONAL PHASE: reconstruction of the private key ###
        secret = (omega1 * alpha1 + omega3 * alpha3) % ec.n
        self.assertIn((q1 + q2 + q3) % ec.n, (secret, ec.n - secret))

    def test_musig(self):
        """testing 3-of-3 MuSig

            https://github.com/ElementsProject/secp256k1-zkp/blob/secp256k1-zkp/src/modules/musig/musig.md
            https://blockstream.com/2019/02/18/musig-a-new-multisignature-standard/
            https://eprint.iacr.org/2018/068
            https://blockstream.com/2018/01/23/musig-key-aggregation-schnorr-signatures.html
            https://medium.com/@snigirev.stepan/how-schnorr-signatures-may-improve-bitcoin-91655bcb4744
        """
        mhd = hf(b'message to sign').digest()

        # the signers private and public keys,
        # including both the curve Point and the BIP340-Schnorr public key
        q1, x_Q1 = ssa.gen_keys()
        x_Q1 = x_Q1.to_bytes(ec.psize, 'big')

        q2, x_Q2 = ssa.gen_keys()
        x_Q2 = x_Q2.to_bytes(ec.psize, 'big')

        q3, x_Q3 = ssa.gen_keys()
        x_Q3 = x_Q3.to_bytes(ec.psize, 'big')

        # (non interactive) key setup
        # this is MuSig core: the rest is just Schnorr signature additivity
        # 1. lexicographic sorting of public keys
        keys: List[bytes] = list()
        keys.append(x_Q1)
        keys.append(x_Q2)
        keys.append(x_Q3)
        keys.sort()
        # 2. coefficients
        prefix = b''.join(keys)
        a1 = int_from_bits(hf(prefix + x_Q1).digest(), ec.nlen) % ec.n
        a2 = int_from_bits(hf(prefix + x_Q2).digest(), ec.nlen) % ec.n
        a3 = int_from_bits(hf(prefix + x_Q3).digest(), ec.nlen) % ec.n
        # 3. aggregated public key
        Q1 = mult(q1)
        Q2 = mult(q2)
        Q3 = mult(q3)
        Q = ec.add(double_mult(a1, Q1, a2, Q2), mult(a3, Q3))
        if not ec.has_square_y(Q):
            # print("Q has been negated")
            a1 = ec.n - a1  # pragma: no cover
            a2 = ec.n - a2  # pragma: no cover
            a3 = ec.n - a3  # pragma: no cover

        # ready to sign: nonces and nonce commitments
        k1, _ = ssa.gen_keys()
        K1 = mult(k1)

        k2, _ = ssa.gen_keys()
        K2 = mult(k2)

        k3, _ = ssa.gen_keys()
        K3 = mult(k3)

        # exchange {K_i} (interactive)

        # computes s_i (non interactive)
        # WARNING: signers must exchange the nonces commitments {K_i}
        # before sharing {s_i}

        # same for all signers
        K = ec.add(ec.add(K1, K2), K3)
        if not ec.has_square_y(K):
            k1 = ec.n - k1  # pragma: no cover
            k2 = ec.n - k2  # pragma: no cover
            k3 = ec.n - k3  # pragma: no cover
        r = K[0]
        e = ssa._challenge(r, Q[0], mhd, ec, hf)
        s1 = (k1 + e * a1 * q1) % ec.n
        s2 = (k2 + e * a2 * q2) % ec.n
        s3 = (k3 + e * a3 * q3) % ec.n

        # exchange s_i (interactive)

        # finalize signature (non interactive)
        s = (s1 + s2 + s3) % ec.n
        sig = r, s
        # check signature is valid
        ssa._verify(mhd, Q, sig, ec, hf)
        self.assertTrue(ssa.verify(mhd, Q, sig))

    def test_crack_prvkey(self):
        q = 0x6A307F426A94F8114701E7C8E774E7F9A47E2C2035DB29A206321725DEADBEEF
        x_Q = mult(q)[0]
        k = 1010101010101010101

        msg1 = "Paolo is afraid of ephemeral random numbers"
        msg1 = hf(msg1.encode()).digest()
        sig1 = ssa.sign(msg1, q, k)
        # print(f'\nmsg1: {msg1.hex().upper()}')
        # print(f'  r1: {hex(sig1[0]).upper()}')
        # print(f'  s1: {hex(sig1[1]).upper()}')

        msg2 = "and Paolo is right to be afraid"
        msg2 = hf(msg2.encode()).digest()
        sig2 = ssa.sign(msg2, q, k)
        # print(f'\nmsg2: {msg2.hex().upper()}')
        # print(f'  r2: {hex(sig2[0]).upper()}')
        # print(f'  s2: {hex(sig2[1]).upper()}')

        qc, kc = ssa.crack_prvkey(msg1, sig1, msg2, sig2, x_Q)
        self.assertIn(q, (qc, ec.n - qc))
        self.assertIn(k, (kc, ec.n - kc))

        self.assertRaises(ValueError, ssa.crack_prvkey, msg1,
                          sig1, msg2, (16, sig1[1]), x_Q)
        self.assertRaises(ValueError, ssa.crack_prvkey,
                          msg1, sig1, msg1, sig1, x_Q)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib import base58address, bech32address, bip32, slip32
from btclib.network import NETWORKS


class TestSLIP32(unittest.TestCase):
    def test_slip32_test_vector(self):
        """SLIP32 test vector

        https://github.com/satoshilabs/slips/blob/master/slip-0132.md
        """
        mnemonic = (
            "abandon abandon abandon abandon abandon abandon "
            "abandon abandon abandon abandon abandon about"
        )
        kpath = "./0/0"
        test_vectors = [
            [
                NETWORKS["mainnet"]["bip32_prv"],
                "m / 44h / 0h / 0h",
                (
                    "xprv9xpXFhFpqdQK3TmytPBqXtGSwS3DLjojFhTGht8gwAAii8py5X6pxeBn"
                    "Q6ehJiyJ6nDjWGJfZ95WxByFXVkDxHXrqu53WCRGypk2ttuqncb"
                ),
                (
                    "xpub6BosfCnifzxcFwrSzQiqu2DBVTshkCXacvNsWGYJVVhhawA7d4R5WSW"
                    "GFNbi8Aw6ZRc1brxMyWMzG3DSSSSoekkudhUd9yLb6qx39T9nMdj"
                ),
                "1LqBGSKuX5yYUonjxT5qGfpUsXKYYWeabA",
            ],
            [
                NETWORKS["mainnet"]["slip32_p2wsh_p2sh_prv"],
                "m / 49h / 0h / 0h",
                (
                    "yprvAHwhK6RbpuS3dgCYHM5jc2ZvEKd7Bi61u9FVhYMpgMSuZS613T1xxQe"
                    "KTffhrHY79hZ5PsskBjcc6C2V7DrnsMsNaGDaWev3GLRQRgV7hxF"
                ),
                (
                    "ypub6Ww3ibxVfGzLrAH1PNcjyAWenMTbbAosGNB6VvmSEgytSER9azLDWCx"
                    "oJwW7Ke7icmizBMXrzBx9979FfaHxHcrArf3zbeJJJUZPf663zsP"
                ),
                "37VucYSaXLCAsxYyAPfbSi9eh4iEcbShgf",
            ],
            [
                NETWORKS["mainnet"]["slip32_p2wpkh_prv"],
                "m / 84h / 0h / 0h",
                (
                    "zprvAdG4iTXWBoARxkkzNpNh8r6Qag3irQB8PzEMkAFeTRXxHpbF9z4QgEv"
                    "BRmfvqWvGp42t42nvgGpNgYSJA9iefm1yYNZKEm7z6qUWCroSQnE"
                ),
                (
                    "zpub6rFR7y4Q2AijBEqTUquhVz398htDFrtymD9xYYfG1m4wAcvPhXNfE3E"
                    "fH1r1ADqtfSdVCToUG868RvUUkgDKf31mGDtKsAYz2oz2AGutZYs"
                ),
                "bc1qcr8te4kr609gcawutmrza0j4xv80jy8z306fyu",
            ],
        ]
        for version, der_path, prv, pub, addr_str in test_vectors:
            rxprv = bip32.mxprv_from_bip39_mnemonic(mnemonic, "", version)
            mxprv = bip32.derive(rxprv, der_path)
            self.assertEqual(prv.encode(), mxprv)
            mxpub = bip32.xpub_from_xprv(mxprv)
            self.assertEqual(pub.encode(), mxpub)
            xpub = bip32.derive(mxpub, kpath)
            address = slip32.address_from_xpub(xpub)
            addr = addr_str.encode()
            self.assertEqual(addr, address)
            address = slip32.address_from_xkey(xpub)
            self.assertEqual(addr, address)
            xprv = bip32.derive(mxprv, kpath)
            address = slip32.address_from_xkey(xprv)
            self.assertEqual(addr, address)
            if version == NETWORKS["mainnet"]["bip32_prv"]:
                address = base58address.p2pkh(xpub)
                self.assertEqual(addr, address)
                address = base58address.p2pkh(xprv)
                self.assertEqual(addr, address)
            elif version == NETWORKS["mainnet"]["slip32_p2wsh_p2sh_prv"]:
                address = base58address.p2wpkh_p2sh(xpub)
                self.assertEqual(addr, address)
                address = base58address.p2wpkh_p2sh(xprv)
                self.assertEqual(addr, address)
            elif version == NETWORKS["mainnet"]["slip32_p2wpkh_prv"]:
                address = bech32address.p2wpkh(xpub)
                self.assertEqual(addr, address)
                address = bech32address.p2wpkh(xprv)
                self.assertEqual(addr, address)

    def test_slip32(self):
        # xkey is not a public one
        xprv = (
            "xprv9s21ZrQH143K2ZP8tyNiUtgoezZosUkw9hhir2JFzDhcUWKz8qFYk3cx"
            "dgSFoCMzt8E2Ubi1nXw71TLhwgCfzqFHfM5Snv4zboSebePRmLS"
        ).encode()
        self.assertRaises(ValueError, slip32.address_from_xpub, xprv)
        address = slip32.address_from_xkey(xprv)
        xpub = bip32.xpub_from_xprv(xprv)
        address2 = slip32.address_from_xpub(xpub)
        self.assertEqual(address, address2)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic curves.

* SEC 2 v.2 curves
  http://www.secg.org/sec2-v2.pdf
* SEC 2 v.1 curves, removed from SEC 2 v.2 as insecure ones
  http://www.secg.org/SEC2-Ver-1.0.pdf
* Federal Information Processing Standards Publication 186-4
  (NIST) curves
  https://oag.ca.gov/sites/all/files/agweb/pdfs/erds1/fips_pub_07_2013.pdf
* Brainpool standard curves
  https://tools.ietf.org/html/rfc5639

"""

import json
from os import path
from typing import Dict

from .curve import Curve

datadir = path.join(path.dirname(__file__), "data")

# Elliptic Curve Cryptography (ECC)
# Brainpool Standard Curves and Curve Generation
# https://tools.ietf.org/html/rfc5639
filename = path.join(datadir, "ec_Brainpool.json")
with open(filename, "r") as f:
    Brainpool_params2 = json.load(f)
Brainpool: Dict[str, Curve] = {}
for ec_name in Brainpool_params2:
    Brainpool[ec_name] = Curve(*Brainpool_params2[ec_name])


# FIPS PUB 186-4
# FEDERAL INFORMATION PROCESSING STANDARDS PUBLICATION
# Digital Signature Standard (DSS)
# https://oag.ca.gov/sites/all/files/agweb/pdfs/erds1/fips_pub_07_2013.pdf
filename = path.join(datadir, "ec_NIST.json")
with open(filename, "r") as f:
    NIST_params2 = json.load(f)
NIST: Dict[str, Curve] = {}
for ec_name in NIST_params2:
    NIST[ec_name] = Curve(*NIST_params2[ec_name])


# SEC 2 v.1 curves, removed from SEC 2 v.2 as insecure ones
# http://www.secg.org/SEC2-Ver-1.0.pdf
filename = path.join(datadir, "ec_SEC2v1_insecure.json")
with open(filename, "r") as f:
    SEC2v1_params2 = json.load(f)
SEC2v1: Dict[str, Curve] = {}
for ec_name in SEC2v1_params2:
    SEC2v1[ec_name] = Curve(*SEC2v1_params2[ec_name])


# curves included in both SEC 2 v.1 and SEC 2 v.2
# http://www.secg.org/sec2-v2.pdf
filename = path.join(datadir, "ec_SEC2v2.json")
with open(filename, "r") as f:
    SEC2v2_params2 = json.load(f)
SEC2v2: Dict[str, Curve] = {}
for ec_name in SEC2v2_params2:
    SEC2v2[ec_name] = Curve(*SEC2v2_params2[ec_name])
    SEC2v1[ec_name] = Curve(*SEC2v2_params2[ec_name])

CURVES = SEC2v1
CURVES.update(NIST)
CURVES.update(Brainpool)

secp256k1 = CURVES["secp256k1"]



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""SLIP32 address.

"""

from . import base58address, bech32address, bip32
from .alias import BIP32Key
from .network import _P2WPKH_P2SH_PUB_PREFIXES, _P2WPKH_PUB_PREFIXES, _XPUB_PREFIXES


def address_from_xkey(xkey: BIP32Key) -> bytes:
    """Return the SLIP32 base58/bech32 address.

    The address is always derived from the compressed public key,
    as this is the default public key representation in BIP32.
    """

    try:
        xkey = bip32.xpub_from_xprv(xkey)
    except Exception:
        pass

    return address_from_xpub(xkey)


def address_from_xpub(xpub: BIP32Key) -> bytes:
    """Return the SLIP32 base58/bech32 address.

    The address is always derived from the compressed public key,
    as this is the default public key representation in BIP32.
    """

    if not isinstance(xpub, dict):
        xpub = bip32.deserialize(xpub)

    if xpub["key"][0] not in (2, 3):
        m = f"Not a public key: {bip32.serialize(xpub).decode()}"
        raise ValueError(m)

    if xpub["version"] in _XPUB_PREFIXES:
        return base58address.p2pkh(xpub)
    elif xpub["version"] in _P2WPKH_PUB_PREFIXES:
        return bech32address.p2wpkh(xpub)
    else:
        # v has been already checked at parsing stage
        # so, v must be in _P2WPKH_P2SH_PUB_PREFIXES
        assert xpub["version"] in _P2WPKH_P2SH_PUB_PREFIXES
        return base58address.p2wpkh_p2sh(xpub)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""SEC compressed/uncompressed point representation."""

from .alias import Octets, Point
from .curve import Curve
from .curves import secp256k1
from .utils import bytes_from_octets


def bytes_from_point(Q: Point, ec: Curve = secp256k1, compressed: bool = True) -> bytes:
    """Return a point as compressed/uncompressed octet sequence.

    Return a point as compressed (0x02, 0x03) or uncompressed (0x04)
    octet sequence, according to SEC 1 v.2, section 2.3.3.
    """

    # check that Q is a point and that is on curve
    ec.require_on_curve(Q)

    if Q[1] == 0:  # infinity point in affine coordinates
        raise ValueError("No bytes representation for the infinity point")

    bPx = Q[0].to_bytes(ec.psize, byteorder="big")
    if compressed:
        return (b"\x03" if (Q[1] & 1) else b"\x02") + bPx

    return b"\x04" + bPx + Q[1].to_bytes(ec.psize, byteorder="big")


def point_from_octets(pubkey: Octets, ec: Curve = secp256k1) -> Point:
    """Return a tuple (Px, Py) that belongs to the curve.

    Return a tuple (Px, Py) that belongs to the curve according to
    SEC 1 v.2, section 2.3.4.
    """

    pubkey = bytes_from_octets(pubkey, (ec.psize + 1, 2 * ec.psize + 1))

    bsize = len(pubkey)  # bytes
    if bsize == ec.psize + 1:  # compressed point
        if pubkey[0] not in (0x02, 0x03):
            msg = f"{ec.psize+1} bytes, "
            msg += f"but not a compressed point prefix ({pubkey[0]})"
            raise ValueError(msg)
        Px = int.from_bytes(pubkey[1:], byteorder="big")
        try:
            Py = ec.y_odd(Px, pubkey[0] % 2)  # also check Px validity
            return Px, Py
        except Exception:
            msg = f"{ec.psize+1} bytes, but not a valid x coordinate {Px}"
            raise ValueError(msg)
    else:  # uncompressed point
        if pubkey[0] != 0x04:
            raise ValueError("not an uncompressed point")
        Px = int.from_bytes(pubkey[1 : ec.psize + 1], byteorder="big")
        P = Px, int.from_bytes(pubkey[ec.psize + 1 :], byteorder="big")
        if P[1] == 0:  # infinity point in affine coordinates
            raise ValueError("No bytes representation for the infinity point")
        if ec.is_on_curve(P):
            return P
        else:
            raise ValueError(f"point {P} not on curve")



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

from typing import Optional

from .alias import PrvKey
from .base58 import b58encode
from .network import NETWORKS
from .to_pubkey import prvkeyinfo_from_prvkey


def wif_from_prvkey(
    prvkey: PrvKey, network: Optional[str] = None, compressed: Optional[bool] = None
) -> bytes:
    "Return the WIF encoding of a private key."

    q, net, compr = prvkeyinfo_from_prvkey(prvkey, network, compressed)
    ec = NETWORKS[net]["curve"]

    payload = NETWORKS[net]["wif"]
    payload += q.to_bytes(ec.nsize, "big")
    payload += b"\x01" if compr else b""
    return b58encode(payload)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic curve classes."""

from math import sqrt
from typing import Union

from .alias import INF, INFJ, Integer, JacPoint, Point
from .numbertheory import legendre_symbol, mod_inv, mod_sqrt
from .utils import int_from_integer


def _jac_from_aff(Q: Point) -> JacPoint:
    # point is assumed to be on curve
    return Q[0], Q[1], 1 if Q[1] else 0


class CurveGroup:
    """Finite group of the points of an elliptic curve over Fp.

    The elliptic curve is the set of points (x, y)
    that are solutions to a Weierstrass equation y^2 = x^3 + a*x + b,
    with x, y, a, and b in Fp (p being a prime),
    together with a point at infinity INF.
    The constants a, b must satisfy the relationship
    4 a^3 + 27 b^2 ≠ 0.

    The group is defined by the point addition group law.
    """

    def __init__(self, p: Integer, a: Integer, b: Integer) -> None:
        # Parameters are checked according to SEC 1 v.2 3.1.1.2.1

        p = int_from_integer(p)
        a = int_from_integer(a)
        b = int_from_integer(b)

        # 1) check that p is a prime
        # Fermat test will do as _probabilistic_ primality test...
        if p < 2 or p % 2 == 0 or pow(2, p - 1, p) != 1:
            raise ValueError(f"p ({hex(p)}) is not prime")

        plen = p.bit_length()
        self.psize = (plen + 7) // 8
        # must be true to break simmetry using quadratic residue
        self.pIsThreeModFour = p % 4 == 3
        self.p = p

        # 2. check that a and b are integers in the interval [0, p−1]
        if not 0 <= a < p:
            raise ValueError(f"invalid a ({hex(a)}) for given p ({hex(p)})")
        if not 0 <= b < p:
            raise ValueError(f"invalid b ({hex(b)}) for given p ({hex(p)})")

        # 3. Check that 4*a^3 + 27*b^2 ≠ 0 (mod p)
        d = 4 * a * a * a + 27 * b * b
        if d % p == 0:
            raise ValueError("zero discriminant")
        self._a = a
        self._b = b

    def __str__(self) -> str:
        result = "Curve"
        result += f"\n p   = {hex(self.p).upper()}"
        result += f"\n a   = {hex(self._a).upper()}"
        result += f"\n b   = {hex(self._b).upper()}"
        return result

    def __repr__(self) -> str:
        result = "Curve("
        result += f"{hex(self.p).upper()}"
        result += f", {hex(self._a).upper()}, {hex(self._b).upper()}"
        result += ")"
        return result

    # methods using p: they could become functions

    def negate(self, Q: Point) -> Point:
        """Return the opposite point on the curve.

        The input point must be on the curve.
        """

        self.require_on_curve(Q)
        # % self.p is required to account for infinity point, i.e. Q[1]==0
        return Q[0], (self.p - Q[1]) % self.p

    def _aff_from_jac(self, Q: JacPoint) -> Point:
        # point is assumed to be on curve
        if Q[2] == 0:  # Infinity point in Jacobian coordinates
            return INF
        else:
            Z2 = Q[2] * Q[2]
            x = (Q[0] * mod_inv(Z2, self.p)) % self.p
            y = (Q[1] * mod_inv(Z2 * Q[2], self.p)) % self.p
            return x, y

    def _x_aff_from_jac(self, Q: JacPoint) -> int:
        # point is assumed to be on curve
        if Q[2] == 0:  # Infinity point in Jacobian coordinates
            raise ValueError("Infinity point has no x-coordinate")
        else:
            Z2 = Q[2] * Q[2]
            return (Q[0] * mod_inv(Z2, self.p)) % self.p

    # methods using _a, _b, _p

    def add(self, Q1: Point, Q2: Point) -> Point:
        """Return the sum of two points.

        The input points must be on the curve.
        """

        self.require_on_curve(Q1)
        self.require_on_curve(Q2)
        # no Jacobian coordinates here as _aff_from_jac would cost 2 mod_inv
        # while _add_aff costs only one mod_inv
        return self._add_aff(Q1, Q2)

    def _add_jac(self, Q: JacPoint, R: JacPoint) -> JacPoint:
        # points are assumed to be on curve

        if Q[2] == 0:  # Infinity point in Jacobian coordinates
            return R
        if R[2] == 0:  # Infinity point in Jacobian coordinates
            return Q

        RZ2 = R[2] * R[2]
        RZ3 = RZ2 * R[2]
        QZ2 = Q[2] * Q[2]
        QZ3 = QZ2 * Q[2]
        if Q[0] * RZ2 % self.p == R[0] * QZ2 % self.p:  # same affine x
            if Q[1] * RZ3 % self.p == R[1] * QZ3 % self.p:  # point doubling
                QY2 = Q[1] * Q[1]
                W = (3 * Q[0] * Q[0] + self._a * QZ2 * QZ2) % self.p
                V = (4 * Q[0] * QY2) % self.p
                X = (W * W - 2 * V) % self.p
                Y = (W * (V - X) - 8 * QY2 * QY2) % self.p
                Z = (2 * Q[1] * Q[2]) % self.p
                return X, Y, Z
            else:  # opposite points
                return INFJ
        else:
            T = (Q[1] * RZ3) % self.p
            U = (R[1] * QZ3) % self.p
            W = (U - T) % self.p

            M = (Q[0] * RZ2) % self.p
            N = (R[0] * QZ2) % self.p
            V = (N - M) % self.p

            V2 = V * V
            V3 = V2 * V
            MV2 = M * V2
            X = (W * W - V3 - 2 * MV2) % self.p
            Y = (W * (MV2 - X) - T * V3) % self.p
            Z = (V * Q[2] * R[2]) % self.p
            return X, Y, Z

    def _add_aff(self, Q: Point, R: Point) -> Point:
        # points are assumed to be on curve
        if R[1] == 0:  # Infinity point in affine coordinates
            return Q
        if Q[1] == 0:  # Infinity point in affine coordinates
            return R

        if R[0] == Q[0]:
            if R[1] == Q[1]:  # point doubling
                lam = (3 * Q[0] * Q[0] + self._a) * mod_inv(2 * Q[1], self.p)
                lam %= self.p
            else:  # opposite points
                return INF
        else:
            lam = ((R[1] - Q[1]) * mod_inv(R[0] - Q[0], self.p)) % self.p
        x = (lam * lam - Q[0] - R[0]) % self.p
        y = (lam * (Q[0] - x) - Q[1]) % self.p
        return x, y

    def _y2(self, x: int) -> int:
        # skipping a crucial check here:
        # if sqrt(y*y) does not exist, then x is not valid.
        # This is a good reason to keep this method private
        return ((x * x + self._a) * x + self._b) % self.p

    def y(self, x: int) -> int:
        """Return the y coordinate from x, as in (x, y)."""
        if not 0 <= x < self.p:
            raise ValueError(f"x-coordinate {hex(x)} not in [0, p-1]")
        y2 = self._y2(x)
        # mod_sqrt will raise a ValueError if root does not exist
        return mod_sqrt(y2, self.p)

    def require_on_curve(self, Q: Point) -> None:
        """Require the input curve Point to be on the curve.

        An Error is raised if not.
        """
        if not self.is_on_curve(Q):
            raise ValueError("Point not on curve")

    def is_on_curve(self, Q: Point) -> bool:
        """Return True if the point is on the curve."""
        if len(Q) != 2:
            raise ValueError("Point must be a tuple[int, int]")
        if Q[1] == 0:  # Infinity point in affine coordinates
            return True
        if not 0 < Q[1] < self.p:  # y cannot be zero
            raise ValueError(f"y-coordinate {hex(Q[1])} not in (0, p)")
        return self._y2(Q[0]) == (Q[1] * Q[1] % self.p)

    def has_square_y(self, Q: Union[Point, JacPoint]) -> bool:
        """Return True if the affine y-coordinate is a square."""
        if len(Q) == 2:
            return legendre_symbol(Q[1], self.p) == 1
        if len(Q) == 3:
            return legendre_symbol(Q[1] * Q[2] % self.p, self.p) == 1
        raise ValueError(f"Not a Point")

    def require_p_ThreeModFour(self) -> None:
        """Require the field prime p to be equal to 3 mod 4.

        An Error is raised if not.
        """
        if not self.pIsThreeModFour:
            m = f"field prime p ({hex(self.p)}) is not equal to 3 (mod 4)"
            raise ValueError(m)

    # break the y simmetry: even/odd, low/high, or quadratic residue criteria

    def y_odd(self, x: int, odd1even0: int = 1) -> int:
        """Return the odd/even affine y-coordinate associated to x."""
        if odd1even0 not in (0, 1):
            raise ValueError("odd1even0 must be bool or 1/0")
        root = self.y(x)
        # switch even/odd root as needed (XORing the conditions)
        return root if root % 2 == odd1even0 else self.p - root

    def y_low(self, x: int, low1high0: int = 1) -> int:
        """Return the low/high affine y-coordinate associated to x."""
        if low1high0 not in (0, 1):
            raise ValueError("low1high0 must be bool or 1/0")
        root = self.y(x)
        # switch low/high root as needed (XORing the conditions)
        return root if (self.p // 2 >= root) == low1high0 else self.p - root

    def y_quadratic_residue(self, x: int, quad_res: int = 1) -> int:
        """Return the quadratic residue affine y-coordinate."""
        if quad_res not in (0, 1):
            raise ValueError("quad_res must be bool or 1/0")
        self.require_p_ThreeModFour()
        root = self.y(x)
        # switch to quadratic residue root as needed
        legendre = legendre_symbol(root, self.p)
        return root if legendre == quad_res else self.p - root


def _mult_aff(m: int, Q: Point, ec: CurveGroup) -> Point:
    # double & add in affine coordinates, using binary decomposition of m
    # Point is assumed to be on curve
    # m is assumed to have been reduced mod n, if appropriate

    if Q[1] == 0:  # Infinity point, affine coordinates
        return INF  # return Infinity point
    R = INF  # initialize as infinity point
    while m > 0:  # use binary representation of m
        if m & 1:  # if least significant bit is 1
            R = ec._add_aff(R, Q)  # then add current Q
        m = m >> 1  # remove the bit just accounted for
        Q = ec._add_aff(Q, Q)  # double Q for next step
    return R


def _mult_jac(m: int, Q: JacPoint, ec: CurveGroup) -> JacPoint:
    # double & add in Jacobian coordinates, using binary decomposition of m
    # Point is assumed to be on curve
    # m is assumed to have been reduced mod n, if appropriate

    if Q[2] == 0:  # Infinity point in affine coordinates
        return INFJ  # return Infinity point
    R = INFJ  # initialize as infinity point
    while m > 0:  # use binary representation of m
        if m & 1:  # if least significant bit is 1
            R = ec._add_jac(R, Q)  # then add current Q
        m = m >> 1  # remove the bit just accounted for
        Q = ec._add_jac(Q, Q)  # double Q for next step
    return R


class CurveSubGroup(CurveGroup):
    "Subgroup of the points of an elliptic curve over Fp generated by G."

    def __init__(self, p: Integer, a: Integer, b: Integer, G: Point) -> None:

        super().__init__(p, a, b)

        # 2. check that xG and yG are integers in the interval [0, p−1]
        # 4. Check that yG^2 = xG^3 + a*xG + b (mod p)
        if len(G) != 2:
            raise ValueError("Generator must a be a sequence[int, int]")
        self.G = (int_from_integer(G[0]), int_from_integer(G[1]))
        if not self.is_on_curve(self.G):
            raise ValueError("Generator is not on the 'x^3 + a*x + b' curve")
        self.GJ = self.G[0], self.G[1], 1  # Jacobian coordinates

    def __str__(self) -> str:
        result = super().__str__()
        result += f"\n x_G = {hex(self.G[0]).upper()}"
        result += f"\n y_G = {hex(self.G[1]).upper()}"
        return result

    def __repr__(self) -> str:
        result = super().__repr__()[:-1]
        result += f", ({hex(self.G[0]).upper()}, {hex(self.G[1]).upper()})"
        result += ")"
        return result


class Curve(CurveSubGroup):
    "Prime order subgroup of the points of an elliptic curve over Fp."

    def __init__(
        self,
        p: Integer,
        a: Integer,
        b: Integer,
        G: Point,
        n: Integer,
        h: int,
        weakness_check: bool = True,
    ) -> None:

        super().__init__(p, a, b, G)
        n = int_from_integer(n)

        # Security level is expressed in bits, where n-bit security
        # means that the attacker would have to perform 2^n operations
        # to break it. Security bits are half the key size for asymmetric
        # elliptic curve cryptography, i.e. half of the number of bits
        # required to express the group order n or, holding Hasse theorem,
        # to express the field prime p

        self.n = n
        self.nlen = n.bit_length()
        self.nsize = (self.nlen + 7) // 8

        # 5. Check that n is prime.
        if n < 2 or n % 2 == 0 or pow(2, n - 1, n) != 1:
            raise ValueError(f"n ({hex(n)}) is not prime")
        delta = int(2 * sqrt(self.p))
        # also check n with Hasse Theorem
        if h < 2:
            if not (self.p + 1 - delta <= n <= self.p + 1 + delta):
                m = f"n ({hex(n)}) not in [p + 1 - delta, p + 1 + delta]"
                raise ValueError(m)

        # 7. Check that G ≠ INF, nG = INF
        if self.G[1] == 0:
            m = "INF point does not generate a prime order subgroup"
            raise ValueError(m)
        Inf = _mult_aff(n, self.G, self)
        if Inf[1] != 0:
            raise ValueError(f"n ({hex(n)}) is not the group order")

        # 6. Check cofactor
        exp_h = int(1 / n + delta / n + self.p / n)
        if h != exp_h:
            raise ValueError(f"h ({h}) not as expected ({exp_h})")
        self.h = h

        # 8. Check that n ≠ p
        assert n != p, f"n=p ({hex(n)}) -> weak curve"
        #    raise UserWarning("n=p -> weak curve")
        if weakness_check:
            # 8. Check that p^i % n ≠ 1 for all 1≤i<100
            for i in range(1, 100):
                if pow(self.p, i, n) == 1:
                    raise UserWarning("weak curve")

    def __str__(self) -> str:
        result = super().__str__()
        result += f"\n n   = {hex(self.n).upper()}"
        result += f"\n h = {self.h}"
        return result

    def __repr__(self) -> str:
        result = super().__repr__()[:-1]
        result += f", {hex(self.n).upper()}"
        result += f", {self.h}"
        result += ")"
        return result



#!/usr/bin/env python3

# Copyright (C) 2019-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import json
import unittest
from hashlib import sha256 as hf
from os import path

from btclib import base58address, bech32address, bip32, bms, dsa
from btclib.base58address import p2pkh, p2wpkh_p2sh
from btclib.base58wif import wif_from_prvkey
from btclib.bech32address import p2wpkh
from btclib.curves import secp256k1 as ec
from btclib.to_prvkey import prvkeyinfo_from_prvkey
from btclib.to_pubkey import pubkeyinfo_from_prvkey


class TestMessageSign(unittest.TestCase):

    def test_signature(self):
        msg = "test message"
        wif, addr = bms.gen_keys()
        sig = bms.sign(msg, wif)
        self.assertTrue(bms.verify(msg, addr, sig))

        # sig taken from (Electrum and) Bitcoin Core
        wif, addr = bms.gen_keys(
            b"5KMWWy2d3Mjc8LojNoj8Lcz9B1aWu8bRofUgGwQk959Dw5h2iyw")
        exp_sig = ("G/iew/NhHV9V9MdUEn/LFOftaTy1ivGPKPKyMlr8OSokNC755fAxpSThN"
                   "RivwTNsyY9vPUDTRYBPc2cmGd5d4y4=").encode()
        sig = bms.sign(msg, wif)
        self.assertEqual(exp_sig, bms.serialize(*sig))
        self.assertEqual(sig, bms.deserialize(sig))
        self.assertTrue(bms.verify(msg, addr, sig))

    def test_msgsign_p2pkh(self):
        msg = "test message"
        # sigs are taken from (Electrum and) Bitcoin Core

        q = "ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb"

        # uncompressed
        wif1u = wif_from_prvkey(q, "mainnet", False)
        self.assertEqual(
            wif1u, b"5KMWWy2d3Mjc8LojNoj8Lcz9B1aWu8bRofUgGwQk959Dw5h2iyw")
        pubkey1u, _ = pubkeyinfo_from_prvkey(wif1u)
        add1u = base58address.p2pkh(pubkey1u)
        self.assertEqual(add1u, b"1HUBHMij46Hae75JPdWjeZ5Q7KaL7EFRSD")
        sig1u = bms.sign(msg, wif1u)
        self.assertTrue(bms.verify(msg, add1u, sig1u))
        self.assertEqual(sig1u[0], 27)
        exp_sig1u = ("G/iew/NhHV9V9MdUEn/LFOftaTy1ivGPKPKyMlr8OSokNC755fAxpST"
                     "hNRivwTNsyY9vPUDTRYBPc2cmGd5d4y4=").encode()
        self.assertEqual(bms.serialize(*sig1u), exp_sig1u)

        # compressed
        wif1c = wif_from_prvkey(q, "mainnet", True)
        self.assertEqual(
            wif1c, b"L41XHGJA5QX43QRG3FEwPbqD5BYvy6WxUxqAMM9oQdHJ5FcRHcGk")
        pubkey1c, _ = pubkeyinfo_from_prvkey(wif1c)
        add1c = base58address.p2pkh(pubkey1c)
        self.assertEqual(add1c, b"14dD6ygPi5WXdwwBTt1FBZK3aD8uDem1FY")
        sig1c = bms.sign(msg, wif1c)
        self.assertTrue(bms.verify(msg, add1c, sig1c))
        self.assertEqual(sig1c[0], 31)
        exp_sig1c = ("H/iew/NhHV9V9MdUEn/LFOftaTy1ivGPKPKyMlr8OSokNC755fAxpST"
                     "hNRivwTNsyY9vPUDTRYBPc2cmGd5d4y4=").encode()
        self.assertEqual(bms.serialize(*sig1c), exp_sig1c)

        self.assertFalse(bms.verify(msg, add1c, sig1u))
        self.assertFalse(bms.verify(msg, add1u, sig1c))

        rf, r, s, = sig1c
        sig1c_malleated_rf = bms.serialize(rf + 1, r, s)
        self.assertFalse(bms.verify(msg, add1c, sig1c_malleated_rf))
        sig1c_malleated_s = bms.serialize(rf, r, ec.n - s)
        self.assertFalse(bms.verify(msg, add1c, sig1c_malleated_s))
        sig1c_malleated_rf_s = bms.serialize(rf + 1, r, ec.n - s)
        self.assertTrue(bms.verify(msg, add1c, sig1c_malleated_rf_s))

    def test_msgsign_p2pkh_2(self):
        msg = "test message"
        # sigs are taken from (Electrum and) Bitcoin Core

        wif = "Ky1XfDK2v6wHPazA6ECaD8UctEoShXdchgABjpU9GWGZDxVRDBMJ"
        # compressed
        address = "1DAag8qiPLHh6hMFVu9qJQm9ro1HtwuyK5"
        exp_sig = ("IFqUo4/sxBEFkfK8mZeeN56V13BqOc0D90oPBChF3gTqMXtNSCTN79UxC"
                   "33kZ8Mi0cHy4zYCnQfCxTyLpMVXKeA=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))
        sig = bms.sign(msg, wif, address)
        self.assertTrue(bms.verify(msg, address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)
        sig = bms.sign(msg.encode(), wif)
        self.assertTrue(bms.verify(msg.encode(), address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

        wif = "5JDopdKaxz5bXVYXcAnfno6oeSL8dpipxtU1AhfKe3Z58X48srn"
        # uncompressed
        address = "19f7adDYqhHSJm2v7igFWZAqxXHj1vUa3T"
        exp_sig = ("HFqUo4/sxBEFkfK8mZeeN56V13BqOc0D90oPBChF3gTqMXtNSCTN79UxC"
                   "33kZ8Mi0cHy4zYCnQfCxTyLpMVXKeA=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))
        sig = bms.sign(msg, wif, address)
        self.assertTrue(bms.verify(msg, address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)
        sig = bms.sign(msg.encode(), wif)
        self.assertTrue(bms.verify(msg.encode(), address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

    def test_verify_p2pkh(self):
        msg = "Hello, world!"
        address = "1FEz167JCVgBvhJBahpzmrsTNewhiwgWVG"
        exp_sig = ("G+WptuOvPCSswt/Ncm1upO4lPSCWbS2cpKariPmHvxX5eOJwgqmdEExMT"
                   "KvaR0S3f1TXwggLn/m4CbI2jv0SCuM=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        # https://github.com/stequald/bitcoin-bms.sign-message
        msg = "test message"
        address = "14dD6ygPi5WXdwwBTt1FBZK3aD8uDem1FY"
        exp_sig = ("IPn9bbEdNUp6+bneZqE2YJbq9Hv5aNILq9E5eZoMSF3/fBX4zjeIN6fpX"
                   "fGSGPrZyKfHQ/c/kTSP+NIwmyTzMfk=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        # https://github.com/stequald/bitcoin-bms.sign-message
        msg = "test message"
        address = "1HUBHMij46Hae75JPdWjeZ5Q7KaL7EFRSD"
        exp_sig = ("G0k+Nt1u5boTTUfLyj6x1T5flg1v9rUKGlhs/jPApaTWLHf3GVdAIOIHi"
                   "p6sVwXEuzQGPWIlS0VT+yryXiDaavw=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        # https://github.com/petertodd/python-bitcoinlib/blob/master/bitcoin/tests/test_signmessage.py
        msg = address = "1F26pNMrywyZJdr22jErtKcjF8R3Ttt55G"
        exp_sig = ("H85WKpqtNZDrajOnYDgUY+abh0KCAcOsAIOQwx2PftAbLEPRA7mzXA/Cj"
                   "XRxzz0MC225pR/hx02Vf2Ag2x33kU4=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        # https://github.com/nanotube/supybot-bitcoin-marketmonitor/blob/master/GPG/local/bitcoinsig.py
        msg = "test message"
        address = "16vqGo3KRKE9kTsTZxKoJKLzwZGTodK3ce"
        exp_sig = ("HPDs1TesA48a9up4QORIuub67VHBM37X66skAYz0Esg23gdfMuCTYDFOR"
                   "c6XGpKZ2/flJ2h/DUF569FJxGoVZ50=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        msg = "test message 2"
        self.assertFalse(bms.verify(msg, address, exp_sig))

        msg = ("freenode:#bitcoin-otc:b42f7e7ea336db4109df6badc05c6b3ea8bfaa13"
               "575b51631c5178a7")
        address = "1GdKjTSg2eMyeVvPV5Nivo6kR8yP2GT7wF"
        exp_sig = ("GyMn9AdYeZIPWLVCiAblOOG18Qqy4fFaqjg5rjH6QT5tNiUXLS6T2o7iu"
                   "WkV1gc4DbEWvyi8yJ8FvSkmEs3voWE=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        msg = "testtest"
        address = "1Hpj6xv9AzaaXjPPisQrdAD2tu84cnPv3f"
        exp_sig = ("INEJxQnSu6mwGnLs0E8eirl5g+0cAC9D5M7hALHD9sK0XQ66CH9mas06g"
                   "NoIX7K1NKTLaj3MzVe8z3pt6apGJ34=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        msg = "testtest"
        address = "18uitB5ARAhyxmkN2Sa9TbEuoGN1he83BX"
        exp_sig = ("IMAtT1SjRyP6bz6vm5tKDTTTNYS6D8w2RQQyKD3VGPq2i2txGd2ar18L8"
                   "/nvF1+kAMo5tNc4x0xAOGP0HRjKLjc=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        msg = "testtest"
        address = "1LsPb3D1o1Z7CzEt1kv5QVxErfqzXxaZXv"
        exp_sig = ("H3I37ur48/fn52ZvWQT+Mj2wXL36gyjfaN5qcgfiVRTJb1eP1li/IacCQ"
                   "spYnUntiRv8r6GDfJYsdiQ5VzlG3As=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        # leading space
        exp_sig = (" H3I37ur48/fn52ZvWQT+Mj2wXL36gyjfaN5qcgfiVRTJb1eP1li/IacC"
                   "QspYnUntiRv8r6GDfJYsdiQ5VzlG3As=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        # trailing space
        exp_sig = ("H3I37ur48/fn52ZvWQT+Mj2wXL36gyjfaN5qcgfiVRTJb1eP1li/IacCQ"
                   "spYnUntiRv8r6GDfJYsdiQ5VzlG3As= ").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        # leading and trailing spaces
        exp_sig = (" H3I37ur48/fn52ZvWQT+Mj2wXL36gyjfaN5qcgfiVRTJb1eP1li/IacC"
                   "QspYnUntiRv8r6GDfJYsdiQ5VzlG3As= ").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

    def test_segwit(self):

        msg = "test"
        wif = "L4xAvhKR35zFcamyHME2ZHfhw5DEyeJvEMovQHQ7DttPTM8NLWCK"
        pubkey, _ = pubkeyinfo_from_prvkey(wif)
        p2pkh = base58address.p2pkh(pubkey)
        p2wpkh = bech32address.p2wpkh(pubkey)
        p2wpkh_p2sh = base58address.p2wpkh_p2sh(pubkey)

        # p2pkh base58 address (Core, Electrum, BIP137)
        exp_sig = ("IBFyn+h9m3pWYbB4fBFKlRzBD4eJKojgCIZSNdhLKKHPSV2/WkeV7R7IO"
                   "I0dpo3uGAEpCz9eepXLrA5kF35MXuU=").encode()
        self.assertTrue(bms.verify(msg, p2pkh, exp_sig))
        sig = bms.sign(msg, wif)  # no address: p2pkh assumed
        self.assertTrue(bms.verify(msg, p2pkh, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

        # p2wpkh-p2sh base58 address (Electrum)
        self.assertTrue(bms.verify(msg, p2wpkh_p2sh, sig))

        # p2wpkh bech32 address (Electrum)
        self.assertTrue(bms.verify(msg, p2wpkh, sig))

        # p2wpkh-p2sh base58 address (BIP137)
        # different first letter in sig because of different rf
        exp_sig = ("JBFyn+h9m3pWYbB4fBFKlRzBD4eJKojgCIZSNdhLKKHPSV2/WkeV7R7IO"
                   "I0dpo3uGAEpCz9eepXLrA5kF35MXuU=").encode()
        self.assertTrue(bms.verify(msg, p2wpkh_p2sh, exp_sig))
        sig = bms.sign(msg, wif, p2wpkh_p2sh)
        self.assertTrue(bms.verify(msg, p2wpkh_p2sh, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

        # p2wpkh bech32 address (BIP137)
        # different first letter in sig because of different rf
        exp_sig = ("KBFyn+h9m3pWYbB4fBFKlRzBD4eJKojgCIZSNdhLKKHPSV2/WkeV7R7IO"
                   "I0dpo3uGAEpCz9eepXLrA5kF35MXuU=").encode()
        self.assertTrue(bms.verify(msg, p2wpkh, exp_sig))
        sig = bms.sign(msg, wif, p2wpkh)
        self.assertTrue(bms.verify(msg, p2wpkh, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

    def test_sign_strippable_message(self):

        wif = "Ky1XfDK2v6wHPazA6ECaD8UctEoShXdchgABjpU9GWGZDxVRDBMJ"
        address = "1DAag8qiPLHh6hMFVu9qJQm9ro1HtwuyK5"

        msg = ""
        exp_sig = ("IFh0InGTy8lLCs03yoUIpJU6MUbi0La/4abhVxyKcCsoUiF3RM7lg51rC"
                   "qyoOZ8Yt43h8LZrmj7nwwO3HIfesiw=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))
        sig = bms.sign(msg, wif)
        self.assertTrue(bms.verify(msg, address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

        # Bitcoin Core exp_sig (Electrum does strip leading/trailing spaces)
        msg = " "
        exp_sig = ("IEveV6CMmOk5lFP+oDbw8cir/OkhJn4S767wt+YwhzHnEYcFOb/uC6rrV"
                   "mTtG3M43mzfObA0Nn1n9CRcv5IGyak=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))
        sig = bms.sign(msg, wif)
        self.assertTrue(bms.verify(msg, address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

        # Bitcoin Core exp_sig (Electrum does strip leading/trailing spaces)
        msg = "  "
        exp_sig = ("H/QjF1V4fVI8IHX8ko0SIypmb0yxfaZLF0o56Cif9z8CX24n4petTxolH"
                   "59pYVMvbTKQkGKpznSiPiQVn83eJF0=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))
        sig = bms.sign(msg, wif)
        self.assertTrue(bms.verify(msg, address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

        msg = "test"
        exp_sig = ("IJUtN/2LZjh1Vx8Ekj9opnIKA6ohKhWB95PLT/3EFgLnOu9hTuYX4+tJJ"
                   "60ZyddFMd6dgAYx15oP+jLw2NzgNUo=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))
        sig = bms.sign(msg, wif)
        self.assertTrue(bms.verify(msg, address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

        # Bitcoin Core exp_sig (Electrum does strip leading/trailing spaces)
        msg = " test "
        exp_sig = ("IA59z13/HBhvMMJtNwT6K7vJByE40lQUdqEMYhX2tnZSD+IGQIoBGE+1I"
                   "YGCHCyqHvTvyGeqJTUx5ywb4StuX0s=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))
        sig = bms.sign(msg, wif)
        self.assertTrue(bms.verify(msg, address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

        # Bitcoin Core exp_sig (Electrum does strip leading/trailing spaces)
        msg = "test "
        exp_sig = ("IPp9l2w0LVYB4FYKBahs+k1/Oa08j+NTuzriDpPWnWQmfU0+UsJNLIPI8"
                   "Q/gekrWPv6sDeYsFSG9VybUKDPGMuo=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))
        sig = bms.sign(msg, wif)
        self.assertTrue(bms.verify(msg, address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

        # Bitcoin Core exp_sig (Electrum does strip leading/trailing spaces)
        msg = " test"
        exp_sig = ("H1nGwD/kcMSmsYU6qihV2l2+Pa+7SPP9zyViZ59VER+QL9cJsIAtu1Cux"
                   "fYDAVt3kgr4t3a/Es3PV82M6z0eQAo=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))
        sig = bms.sign(msg, wif)
        self.assertTrue(bms.verify(msg, address, sig))
        self.assertEqual(bms.serialize(*sig), exp_sig)

    def test_exceptions(self):

        msg = "test"
        wif = "KwELaABegYxcKApCb3kJR9ymecfZZskL9BzVUkQhsqFiUKftb4tu"
        pubkey, _ = pubkeyinfo_from_prvkey(wif)
        address = base58address.p2pkh(pubkey)
        exp_sig = ("IHdKsFF1bUrapA8GMoQUbgI+Ad0ZXyX1c/yAZHmJn5hSNBi7J+TrI1615"
                   "FG3g9JEOPGVvcfDWIFWrg2exLNtoVc=").encode()
        self.assertTrue(bms.verify(msg, address, exp_sig))

        # Invalid recovery flag: 26
        _, r, s = bms.deserialize(exp_sig)
        self.assertRaises(ValueError, bms.serialize, 26, r, s)
        # bms.serialize(26, r, s)

        # short exp_sig
        exp_sig = ("IHdKsFF1bUrapA8GMoQUbgI+Ad0ZXyX1c/yAZHmJn5hNBi7J+TrI1615F"
                   "G3g9JEOPGVvcfDWIFWrg2exLoVc=").encode()
        self.assertRaises(ValueError, bms._verify, msg, address, exp_sig)
        self.assertFalse(bms.verify(msg, address, exp_sig))

        # Invalid recovery flag: 26
        exp_sig = ("GpNLHqEKSzwXV+KwwBfQthQ848mn5qSkmGDXpqshDuPYJELOnSuRYGQQg"
                   "BR4PpI+w2tJdD4v+hxElvAaUSqv2eU=").encode()
        self.assertRaises(ValueError, bms._verify, msg, address, exp_sig)
        self.assertFalse(bms.verify(msg, address, exp_sig))
        # bms._verify(msg, address, exp_sig)

        # Invalid recovery flag: 66
        exp_sig = ("QpNLHqEKSzwXV+KwwBfQthQ848mn5qSkmGDXpqshDuPYJELOnSuRYGQQg"
                   "BR4PpI+w2tJdD4v+hxElvAaUSqv2eU=").encode()
        self.assertRaises(ValueError, bms._verify, msg, address, exp_sig)
        self.assertFalse(bms.verify(msg, address, exp_sig))
        # bms._verify(msg, address, exp_sig)

        # Pubkey mismatch: compressed wif, uncompressed address
        wif = "Ky1XfDK2v6wHPazA6ECaD8UctEoShXdchgABjpU9GWGZDxVRDBMJ"
        address = "19f7adDYqhHSJm2v7igFWZAqxXHj1vUa3T"
        self.assertRaises(ValueError, bms.sign, msg, wif, address)
        # bms.sign(msg, wif, address)

        # Pubkey mismatch: uncompressed wif, compressed address
        wif = "5JDopdKaxz5bXVYXcAnfno6oeSL8dpipxtU1AhfKe3Z58X48srn"
        address = "1DAag8qiPLHh6hMFVu9qJQm9ro1HtwuyK5"
        self.assertRaises(ValueError, bms.sign, msg, wif, address)
        # bms.sign(msg, wif, address)

        msg = "test"
        wif = "L4xAvhKR35zFcamyHME2ZHfhw5DEyeJvEMovQHQ7DttPTM8NLWCK"
        pubkey, _ = pubkeyinfo_from_prvkey(wif)
        p2pkh = base58address.p2pkh(pubkey)
        p2wpkh = bech32address.p2wpkh(pubkey)
        p2wpkh_p2sh = base58address.p2wpkh_p2sh(pubkey)

        wif = "Ky1XfDK2v6wHPazA6ECaD8UctEoShXdchgABjpU9GWGZDxVRDBMJ"
        # Mismatch between p2pkh address and key pair
        self.assertRaises(ValueError, bms.sign, msg, wif, p2pkh)
        # bms.sign(msg, wif, p2pkh)

        # Mismatch between p2wpkh address and key pair
        self.assertRaises(ValueError, bms.sign, msg, wif, p2wpkh)
        # bms.sign(msg, wif, p2wpkh)

        # Mismatch between p2wpkh_p2sh address and key pair
        self.assertRaises(ValueError, bms.sign, msg, wif, p2wpkh_p2sh)
        # bms.sign(msg, wif, p2wpkh_p2sh)

        # Invalid recovery flag (39) for base58 address
        exp_sig = ("IHdKsFF1bUrapA8GMoQUbgI+Ad0ZXyX1c/yAZHmJn5hSNBi7J+TrI1615"
                   "FG3g9JEOPGVvcfDWIFWrg2exLNtoVc=").encode()
        _, r, s = bms.deserialize(exp_sig)
        sig = bms.serialize(39, r, s)
        self.assertRaises(ValueError, bms._verify, msg, p2pkh, sig)
        # bms._verify(msg, p2pkh, sig)

        # Invalid recovery flag (35) for bech32 address
        exp_sig = ("IBFyn+h9m3pWYbB4fBFKlRzBD4eJKojgCIZSNdhLKKHPSV2/WkeV7R7IO"
                   "I0dpo3uGAEpCz9eepXLrA5kF35MXuU=").encode()
        _, r, s = bms.deserialize(exp_sig)
        sig = bms.serialize(35, r, s)
        self.assertRaises(ValueError, bms._verify, msg, p2wpkh, sig)
        # bms._verify(msg, p2wpkh, sig)

    def test_vector_python_bitcoinlib(self):
        """Test python-bitcoinlib test vectors

        https://github.com/petertodd/python-bitcoinlib/blob/master/bitcoin/tests/test_data/bms.json
        """

        fname = "bms.json"
        filename = path.join(path.dirname(__file__), "test_data", fname)
        with open(filename, "r") as f:
            test_vectors = json.load(f)

        for vector in test_vectors[:5]:
            msg = vector["address"]
            tuplesig = bms.sign(msg, vector["wif"])
            self.assertTrue(bms.verify(msg, vector["address"], tuplesig))
            b64sig = bms.serialize(*tuplesig)
            self.assertTrue(bms.verify(msg, vector["address"], b64sig))
            self.assertTrue(bms.verify(
                msg, vector["address"], vector["signature"]))

            # python-bitcoinlib has a signature different from the
            # one generated by Core/Electrum/btclib (which are identical)
            self.assertNotEqual(b64sig.decode(), vector["signature"])

            # python-bitcoinlib does not use RFC6979 deterministic nonce
            # as proved by different r compared to Core/Electrum/btclib
            rf, r, s = tuplesig
            _, r0, _ = bms.deserialize(vector["signature"])
            self.assertNotEqual(r, r0)

            # while Core/Electrum/btclib use "low-s" canonical signature
            self.assertLess(s, ec.n - s)
            # this is not true for python-bitcoinlib
            # self.assertLess(s0, ec.n - s0)
            # self.assertGreater(s0, ec.n - s0)

            # just in case you wonder, here's the malleated signature
            rf += (1 if rf == 31 else -1)
            tuplesig_malleated = rf, r, ec.n - s
            self.assertTrue(bms.verify(
                msg, vector["address"], tuplesig_malleated))
            b64sig_malleated = bms.serialize(*tuplesig_malleated)
            self.assertTrue(bms.verify(
                msg, vector["address"], b64sig_malleated))
            # of course,
            # it is not equal to the python-bitcoinlib one (different r)
            self.assertNotEqual(b64sig_malleated.decode(), vector["signature"])

    def test_ledger(self):
        """Hybrid ECDSA Bitcoin message signature generated by Ledger"""

        mnemonic = ("barely sun    snack   this snack  relief "
                    "pipe   attack disease boss enlist lawsuit")

        # non-standard leading 31 in DER serialization
        derivation_path = "m/1"
        msg = b"\xfb\xa3\x1f\x8cd\x85\xe29#K\xb3{\xfd\xa7<?\x95oL\xee\x19"
        msg += b"\xb2'oh\xa7]\xd9A\xfeU\xd8"
        dersig = ("3144022012ec0c174936c2a46dc657252340b2e6e6dd8c3"
                  "1dd059b6f9f33a90c21af2fba022030e6305b3ccf88009d"
                  "419bf7651afcfcc0a30898b93ae9de9aa6ac03cf8ec56b")

        # pubkey derivation
        rprv = bip32.mxprv_from_bip39_mnemonic(mnemonic)
        xprv = bip32.derive(rprv, derivation_path)
        xpub = bip32.xpub_from_xprv(xprv)

        # the actual message being signed
        magic_msg = bms._magic_hash(msg)

        # save key_id and patch dersig
        dersig = bytes.fromhex(dersig)
        key_id = dersig[0]
        dersig = b"\x30" + dersig[1:]

        r, s = dsa.deserialize(dersig)

        # ECDSA signature verification of the patched dersig
        dsa._verify(magic_msg, xpub, dersig, ec, hf)
        self.assertTrue(dsa.verify(magic_msg, xpub, dersig))

        # compressed address
        addr = base58address.p2pkh(xpub)

        # equivalent Bitcoin Message Signature (non-serialized)
        rec_flag = 27 + 4 + (key_id & 0x01)
        btcmsgsig = (rec_flag, r, s)

        # Bitcoin Message Signature verification
        bms._verify(msg, addr, btcmsgsig)
        self.assertTrue(bms.verify(msg, addr, btcmsgsig))
        self.assertFalse(bms.verify(magic_msg, addr, btcmsgsig))

        bms.sign(msg, xprv)

        # standard leading 30 in DER serialization
        derivation_path = "m/0/0"
        msg = "hello world"
        dersig = ("3045022100967dac3262b4686e89638c8219c5761017f05cd87a855edf"
                  "034f4a3ec6b59d3d0220108a4ef9682b71a45979d8c75c393382d9ccb8"
                  "eb561d73b8c5fc0b87a47e7d27")

        # pubkey derivation
        rprv = bip32.mxprv_from_bip39_mnemonic(mnemonic)
        xprv = bip32.derive(rprv, derivation_path)
        xpub = bip32.xpub_from_xprv(xprv)

        # the actual message being signed
        magic_msg = bms._magic_hash(msg)

        # save key_id and patch dersig
        dersig = bytes.fromhex(dersig)
        key_id = dersig[0]
        dersig = b"\x30" + dersig[1:]

        r, s = dsa.deserialize(dersig)

        # ECDSA signature verification of the patched dersig
        dsa._verify(magic_msg, xpub, dersig, ec, hf)
        self.assertTrue(dsa.verify(magic_msg, xpub, dersig))

        # compressed address
        addr = base58address.p2pkh(xpub)

        # equivalent Bitcoin Message Signature (non-serialized)
        rec_flag = 27 + 4 + (key_id & 0x01)
        btcmsgsig = (rec_flag, r, s)

        # Bitcoin Message Signature verification
        bms._verify(msg, addr, btcmsgsig)
        self.assertTrue(bms.verify(msg, addr, btcmsgsig))
        self.assertFalse(bms.verify(magic_msg, addr, btcmsgsig))

    def test_one_prvkey_multiple_addresses(self):

        msg = "Paolo is afraid of ephemeral random numbers"

        # Compressed WIF
        wif = b"Kx45GeUBSMPReYQwgXiKhG9FzNXrnCeutJp4yjTd5kKxCitadm3C"
        pubkey, network = pubkeyinfo_from_prvkey(wif)
        address1 = p2pkh(pubkey, network)
        address2 = p2wpkh_p2sh(pubkey, network)
        address3 = p2wpkh(pubkey, network)

        # sign with no address (or compressed P2PKH)
        sig1 = bms.sign(msg, wif)
        # True for Bitcoin Core
        self.assertTrue(bms.verify(msg, address1, sig1))
        # True for Electrum p2wpkh_p2sh
        self.assertTrue(bms.verify(msg, address2, sig1))
        # True for Electrum p2wpkh
        self.assertTrue(bms.verify(msg, address3, sig1))

        # sign with p2wpkh_p2sh address (BIP137)
        sig2 = bms.sign(msg, wif, address2)
        # False for Bitcoin Core
        self.assertFalse(bms.verify(msg, address1, sig2))
        # True for BIP137 p2wpkh_p2sh
        self.assertTrue(bms.verify(msg, address2, sig2))
        # False for BIP137 p2wpkh
        self.assertFalse(bms.verify(msg, address3, sig2))

        # sign with p2wpkh address (BIP137)
        sig3 = bms.sign(msg, wif, address3)
        # False for Bitcoin Core
        self.assertFalse(bms.verify(msg, address1, sig3))
        # False for BIP137 p2wpkh_p2sh
        self.assertFalse(bms.verify(msg, address2, sig3))
        # True for BIP137 p2wpkh
        self.assertTrue(bms.verify(msg, address3, sig3))

        # uncompressed WIF / P2PKH address
        q, network, _ = prvkeyinfo_from_prvkey(wif)
        wif2 = wif_from_prvkey(q, network, False)
        pubkey, network = pubkeyinfo_from_prvkey(wif2)
        address4 = p2pkh(pubkey, network)

        # sign with uncompressed P2PKH
        sig4 = bms.sign(msg, wif2, address4)
        # False for Bitcoin Core compressed p2pkh
        self.assertFalse(bms.verify(msg, address1, sig4))
        # False for BIP137 p2wpkh_p2sh
        self.assertFalse(bms.verify(msg, address2, sig4))
        # False for BIP137 p2wpkh
        self.assertFalse(bms.verify(msg, address3, sig4))
        # True for Bitcoin Core uncompressed p2pkh
        self.assertTrue(bms.verify(msg, address4, sig4))

        self.assertRaises(ValueError, bms.sign, msg, wif2, address1)
        self.assertRaises(ValueError, bms.sign, msg, wif, address4)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest
from typing import Dict

from btclib.alias import INF, INFJ
from btclib.curve import Curve, _jac_from_aff
from btclib.curvemult import mult
from btclib.curves import CURVES
from btclib.numbertheory import mod_sqrt
from btclib.secpoint import bytes_from_point, point_from_octets

# test curves: very low cardinality
low_card_curves: Dict[str, Curve] = {}
# 13 % 4 = 1; 13 % 8 = 5
low_card_curves['ec13_11'] = Curve(13, 7, 6, (1, 1), 11, 1, False)
low_card_curves['ec13_19'] = Curve(13, 0, 2, (1, 9), 19, 1, False)
# 17 % 4 = 1; 17 % 8 = 1
low_card_curves['ec17_13'] = Curve(17, 6, 8, (0, 12), 13, 2, False)
low_card_curves['ec17_23'] = Curve(17, 3, 5, (1, 14), 23, 1, False)
# 19 % 4 = 3; 19 % 8 = 3
low_card_curves['ec19_13'] = Curve(19, 0, 2, (4, 16), 13, 2, False)
low_card_curves['ec19_23'] = Curve(19, 2, 9, (0, 16), 23, 1, False)
# 23 % 4 = 3; 23 % 8 = 7
low_card_curves['ec23_19'] = Curve(23, 9, 7, (5, 4), 19, 1, False)
low_card_curves['ec23_31'] = Curve(23, 5, 1, (0, 1), 31, 1, False)

all_curves: Dict[str, Curve] = {}
all_curves.update(low_card_curves)
all_curves.update(CURVES)


class TestEllipticCurve(unittest.TestCase):
    def test_exceptions(self):
        # good
        Curve(11, 2, 7, (6, 9), 7, 2, False)

        # p not odd
        self.assertRaises(ValueError, Curve, 10, 2, 7, (6, 9), 7, 1, False)

        # p not prime
        self.assertRaises(ValueError, Curve, 15, 2, 7, (6, 9), 7, 1, False)

        # a > p
        self.assertRaises(ValueError, Curve, 11, 12, 7, (6, 9), 13, 1, False)

        # b > p
        self.assertRaises(ValueError, Curve, 11, 2, 12, (6, 9), 13, 1, False)

        # zero discriminant
        self.assertRaises(ValueError, Curve, 11, 7, 7, (6, 9), 7, 1, False)

        # G not Tuple (int, int)
        self.assertRaises(ValueError, Curve, 11, 2, 7, (6, 9, 1), 7, 1, False)

        # G not on curve
        self.assertRaises(ValueError, Curve, 11, 2, 7, (7, 9), 7, 1, False)

        # n not prime
        self.assertRaises(ValueError, Curve, 11, 2, 7, (6, 9), 8, 1, False)

        # n not Hesse
        self.assertRaises(ValueError, Curve, 11, 2, 7, (6, 9), 71, 1, True)

        # h not as expected
        self.assertRaises(ValueError, Curve, 11, 2, 7, (6, 9), 7, 1, True)
        # Curve(11, 2, 7, (6, 9), 7, 1, 0, True)

        # n not group order
        self.assertRaises(ValueError, Curve, 11, 2, 7, (6, 9), 13, 1, False)

        # n=p -> weak curve
        # missing

        # weak curve
        self.assertRaises(UserWarning, Curve, 11, 2, 7, (6, 9), 7, 2, True)

        # x-coordinate not in [0, p-1]
        ec = CURVES['secp256k1']
        self.assertRaises(ValueError, ec.y, ec.p)
        # secp256k1.y(secp256k1.p)

        # INF point does not generate a prime order subgroup
        self.assertRaises(ValueError, Curve, 11, 2, 7, INF, 7, 2, False)
        # Curve(11, 2, 7, INF, 7, 2, 0, False)

    def test_all_curves(self):
        for ec in all_curves.values():
            self.assertEqual(mult(0, ec.G, ec), INF)
            self.assertEqual(mult(0, ec.G, ec), INF)

            self.assertEqual(mult(1, ec.G, ec), ec.G)
            self.assertEqual(mult(1, ec.G, ec), ec.G)

            Gy_odd = ec.y_odd(ec.G[0], True)
            self.assertEqual(Gy_odd % 2, 1)
            Gy_even = ec.y_odd(ec.G[0], False)
            self.assertEqual(Gy_even % 2, 0)
            self.assertTrue(ec.G[1] in (Gy_odd, Gy_even))

            Gbytes = bytes_from_point(ec.G, ec)
            G2 = point_from_octets(Gbytes, ec)
            self.assertEqual(ec.G, G2)

            Gbytes = bytes_from_point(ec.G, ec, False)
            G2 = point_from_octets(Gbytes, ec)
            self.assertEqual(ec.G, G2)

            P = ec.add(INF, ec.G)
            self.assertEqual(P, ec.G)
            P = ec.add(ec.G, INF)
            self.assertEqual(P, ec.G)
            P = ec.add(INF, INF)
            self.assertEqual(P, INF)

            P = ec.add(ec.G, ec.G)
            self.assertEqual(P, mult(2, ec.G, ec))

            P = mult(ec.n - 1, ec.G, ec)
            self.assertEqual(ec.add(P, ec.G), INF)
            self.assertEqual(mult(ec.n, ec.G, ec), INF)

            self.assertEqual(mult(0, INF, ec), INF)
            self.assertEqual(mult(1, INF, ec), INF)
            self.assertEqual(mult(25, INF, ec), INF)

            ec_repr = repr(ec)
            if ec in low_card_curves.values() or ec.psize < 24:
                ec_repr = ec_repr[:-1] + ", False)"
            ec2 = eval(ec_repr)
            self.assertEqual(str(ec), str(ec2))

    def test_octets2point(self):
        for ec in all_curves.values():
            Q = mult(ec.p, ec.G, ec)  # just a random point, not INF

            Q_bytes = b'\x03' if Q[1] & 1 else b'\x02'
            Q_bytes += Q[0].to_bytes(ec.psize, byteorder='big')
            R = point_from_octets(Q_bytes, ec)
            self.assertEqual(R, Q)
            self.assertEqual(bytes_from_point(R, ec), Q_bytes)

            Q_hex_str = Q_bytes.hex()
            R = point_from_octets(Q_hex_str, ec)
            self.assertEqual(R, Q)

            Q_bytes = b'\x04' + Q[0].to_bytes(ec.psize, byteorder='big')
            Q_bytes += Q[1].to_bytes(ec.psize, byteorder='big')
            R = point_from_octets(Q_bytes, ec)
            self.assertEqual(R, Q)
            self.assertEqual(bytes_from_point(R, ec, False), Q_bytes)

            Q_hex_str = Q_bytes.hex()
            R = point_from_octets(Q_hex_str, ec)
            self.assertEqual(R, Q)

            # scalar in point multiplication can be int, str, or bytes
            t = tuple()
            self.assertRaises(TypeError, mult, t, ec.G, ec)

            # not a compressed point
            Q_bytes = b'\x01' * (ec.psize + 1)
            self.assertRaises(ValueError, point_from_octets, Q_bytes, ec)
            # not a point
            Q_bytes += b'\x01'
            self.assertRaises(ValueError, point_from_octets, Q_bytes, ec)
            # not an uncompressed point
            Q_bytes = b'\x01' * 2 * (ec.psize + 1)
            self.assertRaises(ValueError, point_from_octets, Q_bytes, ec)

        # invalid x coordinate
        ec = CURVES['secp256k1']
        x = 0xEEFDEA4CDB677750A420FEE807EACF21EB9898AE79B9768766E4FAA04A2D4A34
        xstr = format(x, '32X')
        self.assertRaises(ValueError, point_from_octets, "03" + xstr, ec)
        self.assertRaises(ValueError, point_from_octets, "04" + 2 * xstr, ec)
        self.assertRaises(ValueError, bytes_from_point, (x, x), ec)
        self.assertRaises(ValueError, bytes_from_point, (x, x), ec, False)

        # Point must be a tuple[int, int]
        P = x, x, x
        self.assertRaises(ValueError, ec.is_on_curve, P)

        # y-coordinate not in (0, p)
        P = x, ec.p + 1
        self.assertRaises(ValueError, ec.is_on_curve, P)

    def test_opposite(self):
        for ec in all_curves.values():
            Q = mult(ec.p, ec.G, ec)  # just a random point, not INF
            minus_Q = ec.negate(Q)
            self.assertEqual(ec.add(Q, minus_Q), INF)
            # jacobian coordinates
            Qjac = _jac_from_aff(Q)
            minus_Qjac = _jac_from_aff(minus_Q)
            self.assertEqual(ec._add_jac(Qjac, minus_Qjac)[2], 0)

            # negate of INF is INF
            minus_Inf = ec.negate(INF)
            self.assertEqual(minus_Inf, INF)

    def test_symmetry(self):
        """Methods to break simmetry: quadratic residue, odd/even, low/high"""
        for ec in low_card_curves.values():

            # setup phase
            # compute quadratic residues
            hasRoot = set()
            hasRoot.add(1)

            for i in range(2, ec.p):
                hasRoot.add(i * i % ec.p)

            # test phase
            Q = mult(ec.p, ec.G, ec)  # just a random point, not INF
            x = Q[0]
            if ec.p % 4 == 3:
                quad_res = ec.y_quadratic_residue(x, True)
                not_quad_res = ec.y_quadratic_residue(x, False)
                # in this case only quad_res is a quadratic residue
                self.assertIn(quad_res, hasRoot)
                root = mod_sqrt(quad_res, ec.p)
                self.assertEqual(quad_res, (root * root) % ec.p)
                root = ec.p - root
                self.assertEqual(quad_res, (root * root) % ec.p)

                self.assertTrue(not_quad_res == ec.p - quad_res)
                self.assertNotIn(not_quad_res, hasRoot)
                self.assertRaises(ValueError, mod_sqrt, not_quad_res, ec.p)

                y_odd = ec.y_odd(x, True)
                self.assertTrue(y_odd in (quad_res, not_quad_res))
                self.assertTrue(y_odd % 2 == 1)
                y_even = ec.y_odd(x, False)
                self.assertTrue(y_even in (quad_res, not_quad_res))
                self.assertTrue(y_even % 2 == 0)

                y_low = ec.y_low(x, True)
                self.assertTrue(y_low in (y_odd, y_even))
                y_high = ec.y_low(x, False)
                self.assertTrue(y_high in (y_odd, y_even))
                self.assertTrue(y_low < y_high)
            else:
                self.assertTrue(ec.p % 4 == 1)
                # cannot use y_quadratic_residue in this case
                self.assertRaises(ValueError, ec.y_quadratic_residue, x, True)
                self.assertRaises(ValueError, ec.y_quadratic_residue, x, False)

                y_odd = ec.y_odd(x, True)
                self.assertTrue(y_odd % 2 == 1)
                y_even = ec.y_odd(x, False)
                self.assertTrue(y_even % 2 == 0)
                # in this case neither or both are quadratic residues
                neither = y_odd not in hasRoot and y_even not in hasRoot
                both = y_odd in hasRoot and y_even in hasRoot
                self.assertTrue(neither or both)
                if y_odd in hasRoot:  # both have roots
                    root = mod_sqrt(y_odd, ec.p)
                    self.assertEqual(y_odd, (root * root) % ec.p)
                    root = ec.p - root
                    self.assertEqual(y_odd, (root * root) % ec.p)
                    root = mod_sqrt(y_even, ec.p)
                    self.assertEqual(y_even, (root * root) % ec.p)
                    root = ec.p - root
                    self.assertEqual(y_even, (root * root) % ec.p)
                else:
                    self.assertRaises(ValueError, mod_sqrt, y_odd, ec.p)
                    self.assertRaises(ValueError, mod_sqrt, y_even, ec.p)

                y_low = ec.y_low(x, True)
                self.assertTrue(y_low in (y_odd, y_even))
                y_high = ec.y_low(x, False)
                self.assertTrue(y_high in (y_odd, y_even))
                self.assertTrue(y_low < y_high)

        # with the last curve
        self.assertRaises(ValueError, ec.y_low, x, 2)
        self.assertRaises(ValueError, ec.y_odd, x, 2)
        self.assertRaises(ValueError, ec.y_quadratic_residue, x, 2)

    def test_aff_jac_conversions(self):
        for ec in all_curves.values():
            Q = mult(ec.p, ec.G, ec)  # just a random point, not INF
            QJ = _jac_from_aff(Q)
            checkQ = ec._aff_from_jac(QJ)
            self.assertEqual(Q, checkQ)
            x = ec._x_aff_from_jac(QJ)
            self.assertEqual(Q[0], x)

            checkInf = ec._aff_from_jac(_jac_from_aff(INF))
            self.assertEqual(INF, checkInf)
            # relevant for BIP340-Schnorr signature verification
            self.assertFalse(ec.has_square_y(INF))
            self.assertRaises(ValueError, ec._x_aff_from_jac, INFJ)
            self.assertRaises(ValueError, ec.has_square_y, "Not a Point")

    def test_add(self):
        for ec in all_curves.values():
            Q1 = mult(ec.p, ec.G, ec)  # just a random point, not INF
            Q1J = _jac_from_aff(Q1)

            # distinct points
            Q3 = ec._add_aff(Q1, ec.G)
            Q3jac = ec._add_jac(Q1J, ec.GJ)
            self.assertEqual(Q3, ec._aff_from_jac(Q3jac))

            # point at infinity
            Q3 = ec._add_aff(ec.G, INF)
            Q3jac = ec._add_jac(ec.GJ, INFJ)
            self.assertEqual(Q3, ec._aff_from_jac(Q3jac))
            Q3 = ec._add_aff(INF, ec.G)
            Q3jac = ec._add_jac(INFJ, ec.GJ)
            self.assertEqual(Q3, ec._aff_from_jac(Q3jac))

            # point doubling
            Q3 = ec._add_aff(Q1, Q1)
            Q3jac = ec._add_jac(Q1J, Q1J)
            self.assertEqual(Q3, ec._aff_from_jac(Q3jac))

            # negate points
            Q1opp = ec.negate(Q1)
            Q3 = ec._add_aff(Q1, Q1opp)
            Q3jac = ec._add_jac(Q1J, _jac_from_aff(Q1opp))
            self.assertEqual(Q3, ec._aff_from_jac(Q3jac))


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover
